"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _Arch:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _ArchEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Arch.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    ARCH_UNKNOWN: _Arch.ValueType  # 0
    ARCH_ARM64: _Arch.ValueType  # 1

class Arch(_Arch, metaclass=_ArchEnumTypeWrapper): ...

ARCH_UNKNOWN: Arch.ValueType  # 0
ARCH_ARM64: Arch.ValueType  # 1
global___Arch = Arch

@typing.final
class Command(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MAGIC_FIELD_NUMBER: builtins.int
    EXEC_FIELD_NUMBER: builtins.int
    DLOPEN_FIELD_NUMBER: builtins.int
    DLCLOSE_FIELD_NUMBER: builtins.int
    DLSYM_FIELD_NUMBER: builtins.int
    CALL_FIELD_NUMBER: builtins.int
    PEEK_FIELD_NUMBER: builtins.int
    POKE_FIELD_NUMBER: builtins.int
    LIST_DIR_FIELD_NUMBER: builtins.int
    SHOW_OBJECT_FIELD_NUMBER: builtins.int
    SHOW_CLASS_FIELD_NUMBER: builtins.int
    DUMMY_BLOCK_FIELD_NUMBER: builtins.int
    CLOSE_FIELD_NUMBER: builtins.int
    CLASS_LIST_FIELD_NUMBER: builtins.int
    EXEC_CHUNK_FIELD_NUMBER: builtins.int
    CUSTOM_FIELD_NUMBER: builtins.int
    magic: builtins.int
    @property
    def exec(self) -> global___CmdExec: ...
    @property
    def dlopen(self) -> global___CmdDlopen: ...
    @property
    def dlclose(self) -> global___CmdDlclose: ...
    @property
    def dlsym(self) -> global___CmdDlsym: ...
    @property
    def call(self) -> global___CmdCall: ...
    @property
    def peek(self) -> global___CmdPeek: ...
    @property
    def poke(self) -> global___CmdPoke: ...
    @property
    def list_dir(self) -> global___CmdListDir: ...
    @property
    def show_object(self) -> global___CmdShowObject: ...
    @property
    def show_class(self) -> global___CmdShowClass: ...
    @property
    def dummy_block(self) -> global___CmdDummyBlock: ...
    @property
    def close(self) -> global___CmdClose: ...
    @property
    def class_list(self) -> global___CmdGetClassList: ...
    @property
    def exec_chunk(self) -> global___CmdExecChunk: ...
    @property
    def custom(self) -> global___CmdCustom: ...
    def __init__(
        self,
        *,
        magic: builtins.int = ...,
        exec: global___CmdExec | None = ...,
        dlopen: global___CmdDlopen | None = ...,
        dlclose: global___CmdDlclose | None = ...,
        dlsym: global___CmdDlsym | None = ...,
        call: global___CmdCall | None = ...,
        peek: global___CmdPeek | None = ...,
        poke: global___CmdPoke | None = ...,
        list_dir: global___CmdListDir | None = ...,
        show_object: global___CmdShowObject | None = ...,
        show_class: global___CmdShowClass | None = ...,
        dummy_block: global___CmdDummyBlock | None = ...,
        close: global___CmdClose | None = ...,
        class_list: global___CmdGetClassList | None = ...,
        exec_chunk: global___CmdExecChunk | None = ...,
        custom: global___CmdCustom | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["call", b"call", "class_list", b"class_list", "close", b"close", "custom", b"custom", "dlclose", b"dlclose", "dlopen", b"dlopen", "dlsym", b"dlsym", "dummy_block", b"dummy_block", "exec", b"exec", "exec_chunk", b"exec_chunk", "list_dir", b"list_dir", "peek", b"peek", "poke", b"poke", "show_class", b"show_class", "show_object", b"show_object", "type", b"type"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["call", b"call", "class_list", b"class_list", "close", b"close", "custom", b"custom", "dlclose", b"dlclose", "dlopen", b"dlopen", "dlsym", b"dlsym", "dummy_block", b"dummy_block", "exec", b"exec", "exec_chunk", b"exec_chunk", "list_dir", b"list_dir", "magic", b"magic", "peek", b"peek", "poke", b"poke", "show_class", b"show_class", "show_object", b"show_object", "type", b"type"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["type", b"type"]) -> typing.Literal["exec", "dlopen", "dlclose", "dlsym", "call", "peek", "poke", "list_dir", "show_object", "show_class", "dummy_block", "close", "class_list", "exec_chunk", "custom"] | None: ...

global___Command = Command

@typing.final
class Response(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EXEC_FIELD_NUMBER: builtins.int
    EXEC_CHUNK_FIELD_NUMBER: builtins.int
    DLOPEN_FIELD_NUMBER: builtins.int
    DLCLOSE_FIELD_NUMBER: builtins.int
    DLSYM_FIELD_NUMBER: builtins.int
    PEEK_FIELD_NUMBER: builtins.int
    POKE_FIELD_NUMBER: builtins.int
    CALL_FIELD_NUMBER: builtins.int
    ERROR_FIELD_NUMBER: builtins.int
    DUMMY_BLOCK_FIELD_NUMBER: builtins.int
    SHOW_OBJECT_FIELD_NUMBER: builtins.int
    CLASS_LIST_FIELD_NUMBER: builtins.int
    SHOW_CLASS_FIELD_NUMBER: builtins.int
    LIST_DIR_FIELD_NUMBER: builtins.int
    CUSTOM_FIELD_NUMBER: builtins.int
    @property
    def exec(self) -> global___ResponseCmdExec: ...
    @property
    def exec_chunk(self) -> global___ResponseCmdExecChunk: ...
    @property
    def dlopen(self) -> global___ResponseDlopen: ...
    @property
    def dlclose(self) -> global___ResponseDlclose: ...
    @property
    def dlsym(self) -> global___ResponseDlsym: ...
    @property
    def peek(self) -> global___ResponsePeek: ...
    @property
    def poke(self) -> global___ResponsePoke: ...
    @property
    def call(self) -> global___ResponseCall: ...
    @property
    def error(self) -> global___ResponseError: ...
    @property
    def dummy_block(self) -> global___ResponseDummyBlock: ...
    @property
    def show_object(self) -> global___ResponseShowObject: ...
    @property
    def class_list(self) -> global___ResponseGetClassList: ...
    @property
    def show_class(self) -> global___ResponseShowClass: ...
    @property
    def list_dir(self) -> global___ResponseListdir: ...
    @property
    def custom(self) -> global___ResponseCustom: ...
    def __init__(
        self,
        *,
        exec: global___ResponseCmdExec | None = ...,
        exec_chunk: global___ResponseCmdExecChunk | None = ...,
        dlopen: global___ResponseDlopen | None = ...,
        dlclose: global___ResponseDlclose | None = ...,
        dlsym: global___ResponseDlsym | None = ...,
        peek: global___ResponsePeek | None = ...,
        poke: global___ResponsePoke | None = ...,
        call: global___ResponseCall | None = ...,
        error: global___ResponseError | None = ...,
        dummy_block: global___ResponseDummyBlock | None = ...,
        show_object: global___ResponseShowObject | None = ...,
        class_list: global___ResponseGetClassList | None = ...,
        show_class: global___ResponseShowClass | None = ...,
        list_dir: global___ResponseListdir | None = ...,
        custom: global___ResponseCustom | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["call", b"call", "class_list", b"class_list", "custom", b"custom", "dlclose", b"dlclose", "dlopen", b"dlopen", "dlsym", b"dlsym", "dummy_block", b"dummy_block", "error", b"error", "exec", b"exec", "exec_chunk", b"exec_chunk", "list_dir", b"list_dir", "peek", b"peek", "poke", b"poke", "show_class", b"show_class", "show_object", b"show_object", "type", b"type"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["call", b"call", "class_list", b"class_list", "custom", b"custom", "dlclose", b"dlclose", "dlopen", b"dlopen", "dlsym", b"dlsym", "dummy_block", b"dummy_block", "error", b"error", "exec", b"exec", "exec_chunk", b"exec_chunk", "list_dir", b"list_dir", "peek", b"peek", "poke", b"poke", "show_class", b"show_class", "show_object", b"show_object", "type", b"type"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["type", b"type"]) -> typing.Literal["exec", "exec_chunk", "dlopen", "dlclose", "dlsym", "peek", "poke", "call", "error", "dummy_block", "show_object", "class_list", "show_class", "list_dir", "custom"] | None: ...

global___Response = Response

@typing.final
class ReturnRegistersArm(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    X0_FIELD_NUMBER: builtins.int
    X1_FIELD_NUMBER: builtins.int
    X2_FIELD_NUMBER: builtins.int
    X3_FIELD_NUMBER: builtins.int
    X4_FIELD_NUMBER: builtins.int
    X5_FIELD_NUMBER: builtins.int
    X6_FIELD_NUMBER: builtins.int
    X7_FIELD_NUMBER: builtins.int
    D0_FIELD_NUMBER: builtins.int
    D1_FIELD_NUMBER: builtins.int
    D2_FIELD_NUMBER: builtins.int
    D3_FIELD_NUMBER: builtins.int
    D4_FIELD_NUMBER: builtins.int
    D5_FIELD_NUMBER: builtins.int
    D6_FIELD_NUMBER: builtins.int
    D7_FIELD_NUMBER: builtins.int
    x0: builtins.int
    x1: builtins.int
    x2: builtins.int
    x3: builtins.int
    x4: builtins.int
    x5: builtins.int
    x6: builtins.int
    x7: builtins.int
    d0: builtins.float
    d1: builtins.float
    d2: builtins.float
    d3: builtins.float
    d4: builtins.float
    d5: builtins.float
    d6: builtins.float
    d7: builtins.float
    def __init__(
        self,
        *,
        x0: builtins.int = ...,
        x1: builtins.int = ...,
        x2: builtins.int = ...,
        x3: builtins.int = ...,
        x4: builtins.int = ...,
        x5: builtins.int = ...,
        x6: builtins.int = ...,
        x7: builtins.int = ...,
        d0: builtins.float = ...,
        d1: builtins.float = ...,
        d2: builtins.float = ...,
        d3: builtins.float = ...,
        d4: builtins.float = ...,
        d5: builtins.float = ...,
        d6: builtins.float = ...,
        d7: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["d0", b"d0", "d1", b"d1", "d2", b"d2", "d3", b"d3", "d4", b"d4", "d5", b"d5", "d6", b"d6", "d7", b"d7", "x0", b"x0", "x1", b"x1", "x2", b"x2", "x3", b"x3", "x4", b"x4", "x5", b"x5", "x6", b"x6", "x7", b"x7"]) -> None: ...

global___ReturnRegistersArm = ReturnRegistersArm

@typing.final
class Argument(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    V_INT_FIELD_NUMBER: builtins.int
    V_DOUBLE_FIELD_NUMBER: builtins.int
    V_STR_FIELD_NUMBER: builtins.int
    V_BYTES_FIELD_NUMBER: builtins.int
    v_int: builtins.int
    v_double: builtins.float
    v_str: builtins.str
    v_bytes: builtins.bytes
    def __init__(
        self,
        *,
        v_int: builtins.int = ...,
        v_double: builtins.float = ...,
        v_str: builtins.str = ...,
        v_bytes: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["type", b"type", "v_bytes", b"v_bytes", "v_double", b"v_double", "v_int", b"v_int", "v_str", b"v_str"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["type", b"type", "v_bytes", b"v_bytes", "v_double", b"v_double", "v_int", b"v_int", "v_str", b"v_str"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["type", b"type"]) -> typing.Literal["v_int", "v_double", "v_str", "v_bytes"] | None: ...

global___Argument = Argument

@typing.final
class Handshake(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MAGIC_FIELD_NUMBER: builtins.int
    ARCH_FIELD_NUMBER: builtins.int
    SYSNAME_FIELD_NUMBER: builtins.int
    MACHINE_FIELD_NUMBER: builtins.int
    magic: builtins.int
    arch: global___Arch.ValueType
    sysname: builtins.str
    machine: builtins.str
    def __init__(
        self,
        *,
        magic: builtins.int = ...,
        arch: global___Arch.ValueType = ...,
        sysname: builtins.str = ...,
        machine: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["arch", b"arch", "machine", b"machine", "magic", b"magic", "sysname", b"sysname"]) -> None: ...

global___Handshake = Handshake

@typing.final
class ObjcClass(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ADDRESS_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    address: builtins.int
    name: builtins.str
    def __init__(
        self,
        *,
        address: builtins.int = ...,
        name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["address", b"address", "name", b"name"]) -> None: ...

global___ObjcClass = ObjcClass

@typing.final
class CmdShowObject(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ADDRESS_FIELD_NUMBER: builtins.int
    address: builtins.int
    def __init__(
        self,
        *,
        address: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["address", b"address"]) -> None: ...

global___CmdShowObject = CmdShowObject

@typing.final
class ResponseShowObject(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DESCRIPTION_FIELD_NUMBER: builtins.int
    description: builtins.str
    def __init__(
        self,
        *,
        description: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["description", b"description"]) -> None: ...

global___ResponseShowObject = ResponseShowObject

@typing.final
class CmdShowClass(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ADDRESS_FIELD_NUMBER: builtins.int
    address: builtins.int
    def __init__(
        self,
        *,
        address: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["address", b"address"]) -> None: ...

global___CmdShowClass = CmdShowClass

@typing.final
class ResponseShowClass(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DESCRIPTION_FIELD_NUMBER: builtins.int
    description: builtins.str
    def __init__(
        self,
        *,
        description: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["description", b"description"]) -> None: ...

global___ResponseShowClass = ResponseShowClass

@typing.final
class CmdExecChunk(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BUFFER_FIELD_NUMBER: builtins.int
    buffer: builtins.bytes
    def __init__(
        self,
        *,
        buffer: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["buffer", b"buffer"]) -> None: ...

global___CmdExecChunk = CmdExecChunk

@typing.final
class ResponseCmdExecChunk(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BUFFER_FIELD_NUMBER: builtins.int
    EXIT_CODE_FIELD_NUMBER: builtins.int
    buffer: builtins.bytes
    exit_code: builtins.int
    def __init__(
        self,
        *,
        buffer: builtins.bytes = ...,
        exit_code: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["buffer", b"buffer", "exit_code", b"exit_code", "type", b"type"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["buffer", b"buffer", "exit_code", b"exit_code", "type", b"type"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["type", b"type"]) -> typing.Literal["buffer", "exit_code"] | None: ...

global___ResponseCmdExecChunk = ResponseCmdExecChunk

@typing.final
class CmdDlopen(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILENAME_FIELD_NUMBER: builtins.int
    MODE_FIELD_NUMBER: builtins.int
    filename: builtins.str
    mode: builtins.int
    def __init__(
        self,
        *,
        filename: builtins.str = ...,
        mode: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["filename", b"filename", "mode", b"mode"]) -> None: ...

global___CmdDlopen = CmdDlopen

@typing.final
class ResponseDlopen(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    HANDLE_FIELD_NUMBER: builtins.int
    handle: builtins.int
    def __init__(
        self,
        *,
        handle: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["handle", b"handle"]) -> None: ...

global___ResponseDlopen = ResponseDlopen

@typing.final
class CmdDlclose(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    HANDLE_FIELD_NUMBER: builtins.int
    handle: builtins.int
    def __init__(
        self,
        *,
        handle: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["handle", b"handle"]) -> None: ...

global___CmdDlclose = CmdDlclose

@typing.final
class ResponseDlclose(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RES_FIELD_NUMBER: builtins.int
    res: builtins.int
    def __init__(
        self,
        *,
        res: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["res", b"res"]) -> None: ...

global___ResponseDlclose = ResponseDlclose

@typing.final
class CmdDlsym(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    HANDLE_FIELD_NUMBER: builtins.int
    SYMBOL_NAME_FIELD_NUMBER: builtins.int
    handle: builtins.int
    symbol_name: builtins.str
    def __init__(
        self,
        *,
        handle: builtins.int = ...,
        symbol_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["handle", b"handle", "symbol_name", b"symbol_name"]) -> None: ...

global___CmdDlsym = CmdDlsym

@typing.final
class ResponseDlsym(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PTR_FIELD_NUMBER: builtins.int
    ptr: builtins.int
    def __init__(
        self,
        *,
        ptr: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["ptr", b"ptr"]) -> None: ...

global___ResponseDlsym = ResponseDlsym

@typing.final
class CmdExec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BACKGROUND_FIELD_NUMBER: builtins.int
    ARGV_FIELD_NUMBER: builtins.int
    ENVP_FIELD_NUMBER: builtins.int
    background: builtins.bool
    @property
    def argv(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def envp(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        background: builtins.bool = ...,
        argv: collections.abc.Iterable[builtins.str] | None = ...,
        envp: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["argv", b"argv", "background", b"background", "envp", b"envp"]) -> None: ...

global___CmdExec = CmdExec

@typing.final
class ResponseCmdExec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PID_FIELD_NUMBER: builtins.int
    pid: builtins.int
    def __init__(
        self,
        *,
        pid: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["pid", b"pid"]) -> None: ...

global___ResponseCmdExec = ResponseCmdExec

@typing.final
class CmdCall(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ADDRESS_FIELD_NUMBER: builtins.int
    VA_LIST_INDEX_FIELD_NUMBER: builtins.int
    ARGV_FIELD_NUMBER: builtins.int
    address: builtins.int
    va_list_index: builtins.int
    @property
    def argv(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Argument]: ...
    def __init__(
        self,
        *,
        address: builtins.int = ...,
        va_list_index: builtins.int = ...,
        argv: collections.abc.Iterable[global___Argument] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["address", b"address", "argv", b"argv", "va_list_index", b"va_list_index"]) -> None: ...

global___CmdCall = CmdCall

@typing.final
class ResponseCall(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ARM_REGISTERS_FIELD_NUMBER: builtins.int
    RETURN_VALUE_FIELD_NUMBER: builtins.int
    return_value: builtins.int
    @property
    def arm_registers(self) -> global___ReturnRegistersArm: ...
    def __init__(
        self,
        *,
        arm_registers: global___ReturnRegistersArm | None = ...,
        return_value: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["arm_registers", b"arm_registers", "return_value", b"return_value", "return_values", b"return_values"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["arm_registers", b"arm_registers", "return_value", b"return_value", "return_values", b"return_values"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["return_values", b"return_values"]) -> typing.Literal["arm_registers", "return_value"] | None: ...

global___ResponseCall = ResponseCall

@typing.final
class CmdPeek(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ADDRESS_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    address: builtins.int
    size: builtins.int
    def __init__(
        self,
        *,
        address: builtins.int = ...,
        size: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["address", b"address", "size", b"size"]) -> None: ...

global___CmdPeek = CmdPeek

@typing.final
class ResponsePeek(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATA_FIELD_NUMBER: builtins.int
    data: builtins.bytes
    def __init__(
        self,
        *,
        data: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["data", b"data"]) -> None: ...

global___ResponsePeek = ResponsePeek

@typing.final
class CmdPoke(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ADDRESS_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    address: builtins.int
    data: builtins.bytes
    def __init__(
        self,
        *,
        address: builtins.int = ...,
        data: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["address", b"address", "data", b"data"]) -> None: ...

global___CmdPoke = CmdPoke

@typing.final
class ResponsePoke(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___ResponsePoke = ResponsePoke

@typing.final
class CmdListDir(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PATH_FIELD_NUMBER: builtins.int
    path: builtins.str
    def __init__(
        self,
        *,
        path: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["path", b"path"]) -> None: ...

global___CmdListDir = CmdListDir

@typing.final
class CmdDummyBlock(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___CmdDummyBlock = CmdDummyBlock

@typing.final
class ResponseDummyBlock(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ADDRESS_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    address: builtins.int
    size: builtins.int
    def __init__(
        self,
        *,
        address: builtins.int = ...,
        size: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["address", b"address", "size", b"size"]) -> None: ...

global___ResponseDummyBlock = ResponseDummyBlock

@typing.final
class CmdGetClassList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___CmdGetClassList = CmdGetClassList

@typing.final
class ResponseGetClassList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CLASSES_FIELD_NUMBER: builtins.int
    @property
    def classes(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ObjcClass]: ...
    def __init__(
        self,
        *,
        classes: collections.abc.Iterable[global___ObjcClass] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["classes", b"classes"]) -> None: ...

global___ResponseGetClassList = ResponseGetClassList

@typing.final
class ResponseError(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___ResponseError = ResponseError

@typing.final
class ResponseListdir(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MAGIC_FIELD_NUMBER: builtins.int
    DIRP_FIELD_NUMBER: builtins.int
    DIR_ENTRIES_FIELD_NUMBER: builtins.int
    magic: builtins.int
    dirp: builtins.int
    @property
    def dir_entries(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___DirEntry]: ...
    def __init__(
        self,
        *,
        magic: builtins.int = ...,
        dirp: builtins.int = ...,
        dir_entries: collections.abc.Iterable[global___DirEntry] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["dir_entries", b"dir_entries", "dirp", b"dirp", "magic", b"magic"]) -> None: ...

global___ResponseListdir = ResponseListdir

@typing.final
class DirEntry(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    D_TYPE_FIELD_NUMBER: builtins.int
    D_NAME_FIELD_NUMBER: builtins.int
    LSTAT_FIELD_NUMBER: builtins.int
    STAT_FIELD_NUMBER: builtins.int
    d_type: builtins.int
    d_name: builtins.str
    @property
    def lstat(self) -> global___DirEntryStat: ...
    @property
    def stat(self) -> global___DirEntryStat: ...
    def __init__(
        self,
        *,
        d_type: builtins.int = ...,
        d_name: builtins.str = ...,
        lstat: global___DirEntryStat | None = ...,
        stat: global___DirEntryStat | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["lstat", b"lstat", "stat", b"stat"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["d_name", b"d_name", "d_type", b"d_type", "lstat", b"lstat", "stat", b"stat"]) -> None: ...

global___DirEntry = DirEntry

@typing.final
class DirEntryStat(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ERRNO1_FIELD_NUMBER: builtins.int
    ST_DEV_FIELD_NUMBER: builtins.int
    ST_MODE_FIELD_NUMBER: builtins.int
    ST_NLINK_FIELD_NUMBER: builtins.int
    ST_INO_FIELD_NUMBER: builtins.int
    ST_UID_FIELD_NUMBER: builtins.int
    ST_GID_FIELD_NUMBER: builtins.int
    ST_RDEV_FIELD_NUMBER: builtins.int
    ST_SIZE_FIELD_NUMBER: builtins.int
    ST_BLOCKS_FIELD_NUMBER: builtins.int
    ST_BLKSIZE_FIELD_NUMBER: builtins.int
    ST_ATIME1_FIELD_NUMBER: builtins.int
    ST_MTIME1_FIELD_NUMBER: builtins.int
    ST_CTIME1_FIELD_NUMBER: builtins.int
    errno1: builtins.int
    st_dev: builtins.int
    st_mode: builtins.int
    st_nlink: builtins.int
    st_ino: builtins.int
    st_uid: builtins.int
    st_gid: builtins.int
    st_rdev: builtins.int
    st_size: builtins.int
    st_blocks: builtins.int
    st_blksize: builtins.int
    st_atime1: builtins.int
    st_mtime1: builtins.int
    st_ctime1: builtins.int
    def __init__(
        self,
        *,
        errno1: builtins.int = ...,
        st_dev: builtins.int = ...,
        st_mode: builtins.int = ...,
        st_nlink: builtins.int = ...,
        st_ino: builtins.int = ...,
        st_uid: builtins.int = ...,
        st_gid: builtins.int = ...,
        st_rdev: builtins.int = ...,
        st_size: builtins.int = ...,
        st_blocks: builtins.int = ...,
        st_blksize: builtins.int = ...,
        st_atime1: builtins.int = ...,
        st_mtime1: builtins.int = ...,
        st_ctime1: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["errno1", b"errno1", "st_atime1", b"st_atime1", "st_blksize", b"st_blksize", "st_blocks", b"st_blocks", "st_ctime1", b"st_ctime1", "st_dev", b"st_dev", "st_gid", b"st_gid", "st_ino", b"st_ino", "st_mode", b"st_mode", "st_mtime1", b"st_mtime1", "st_nlink", b"st_nlink", "st_rdev", b"st_rdev", "st_size", b"st_size", "st_uid", b"st_uid"]) -> None: ...

global___DirEntryStat = DirEntryStat

@typing.final
class CmdClose(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___CmdClose = CmdClose

@typing.final
class CmdCustom(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONTENT_FIELD_NUMBER: builtins.int
    content: builtins.bytes
    def __init__(
        self,
        *,
        content: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["content", b"content"]) -> None: ...

global___CmdCustom = CmdCustom

@typing.final
class ResponseCustom(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONTENT_FIELD_NUMBER: builtins.int
    content: builtins.bytes
    def __init__(
        self,
        *,
        content: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["content", b"content"]) -> None: ...

global___ResponseCustom = ResponseCustom



from typing import Optional

from rpcclient.exceptions import MissingLibraryError
from rpcclient.structs.consts import RTLD_NOW


class Bluetooth:
    """ bluetooth utils """

    # bugfix: +[BluetoothManager.setSharedInstanceQueue:] cannot be called twice in the same process,
    # so we use this global to tell if it was already called
    _ENV_QUEUE_SET = '_rpc_server_bluetooth_manager_dispatch_queue_set'

    def __init__(self, client):
        self._client = client
        self._load_bluetooth_manager()

        bluetooth_manager_class = client.symbols.objc_getClass('BluetoothManager')
        if not client.getenv(self._ENV_QUEUE_SET):
            bluetooth_manager_class.objc_call('setSharedInstanceQueue:',
                                              self._client.symbols.dispatch_queue_create(0, 0))
            client.setenv(self._ENV_QUEUE_SET, '1')
        self._bluetooth_manager = bluetooth_manager_class.objc_call('sharedInstance')

    def is_on(self):
        return 1 == self._bluetooth_manager.objc_call('enabled')

    def turn_on(self):
        self._set(is_on=1)

    def turn_off(self):
        self._set(is_on=0)

    @property
    def address(self) -> Optional[str]:
        return self._bluetooth_manager.objc_call('localAddress').py()

    @property
    def connected(self) -> bool:
        return bool(self._bluetooth_manager.objc_call('connected'))

    @property
    def discoverable(self) -> bool:
        return bool(self._bluetooth_manager.objc_call('isDiscoverable'))

    @discoverable.setter
    def discoverable(self, value: bool):
        self._bluetooth_manager.objc_call('setDiscoverable:', value)

    def _load_bluetooth_manager(self):
        options = [
            '/System/Library/PrivateFrameworks/BluetoothManager.framework/BluetoothManager',
        ]
        for option in options:
            if self._client.dlopen(option, RTLD_NOW):
                return
        raise MissingLibraryError('failed to load BluetoothManager')

    def _set(self, is_on):
        self._bluetooth_manager.objc_call('setPowered:', is_on)
        self._bluetooth_manager.objc_call('setEnabled:', is_on)

    def __repr__(self):
        return f"<Bluetooth state:{'ON' if self.is_on() else 'OFF'}>"



from contextlib import closing
from uuid import uuid4

import pytest

from rpcclient.client_factory import create_tcp_client
from rpcclient.darwin.client import DarwinClient
from rpcclient.exceptions import BadReturnValueError
from rpcclient.ios.client import IosClient
from rpcclient.protos.rpc_pb2 import ARCH_ARM64


@pytest.fixture
def client():
    with closing(create_tcp_client('127.0.0.1')) as c:
        yield c


@pytest.fixture
def tmp_path(client):
    tmp_path = '/tmp'
    try:
        tmp_path = client.fs.readlink(tmp_path)
    except BadReturnValueError:
        pass
    filename = client.fs.remote_path(tmp_path) / uuid4().hex
    client.fs.mkdir(filename, mode=0o777)
    try:
        yield filename
    finally:
        client.fs.remove(filename, recursive=True)


def pytest_addoption(parser):
    parser.addoption('--ci', action='store_true', default=False, help='Don\'t run local only tests')
    parser.addoption('--local-machine', action='store_true', default=False, help='Run local-machine tests')


def pytest_configure(config):
    config.addinivalue_line(
        'markers',
        '''local_only: marks tests that require features the CI lacks (deselect with '-m "not local_only"')'''
    )
    config.addinivalue_line('markers', 'darwin: marks tests that require darwin platform to run')
    config.addinivalue_line('markers', 'ios: marks tests that require ios platform to run')
    config.addinivalue_line('markers', 'local_machine: marks tests that require local_machine to run')
    config.addinivalue_line('markers', 'arm: marks tests that require arm architecture to run')


def pytest_collection_modifyitems(config, items):
    skip_local_only = pytest.mark.skip(reason='remove --ci option to run')
    skip_not_darwin = pytest.mark.skip(reason='Darwin system is required for this test')
    skip_not_ios = pytest.mark.skip(reason='Ios system is required for this test')
    skip_not_arm = pytest.mark.skip(reason='Arm arch is required for this test')
    skip_not_local_machine = pytest.mark.skip(reason='Local machine is required for this test')

    with closing(create_tcp_client('127.0.0.1')) as c:
        is_darwin = isinstance(c, DarwinClient)
        is_ios = isinstance(c, IosClient)
        is_arm = c.arch == ARCH_ARM64

    for item in items:
        if 'local_only' in item.keywords and config.getoption('--ci'):
            # --ci given in cli: skip local only tests.
            item.add_marker(skip_local_only)
        if 'darwin' in item.keywords and not is_darwin:
            # Skip test that require Darwin on non Darwin system
            item.add_marker(skip_not_darwin)
        if 'ios' in item.keywords and not is_ios:
            # Skip test that require ios
            item.add_marker(skip_not_ios)
        if 'arm' in item.keywords and not is_arm:
            # Skip tests that require arm on non arm architecture
            item.add_marker(skip_not_arm)
        if 'local_machine' in item.keywords and not config.getoption('--local-machine'):
            # Skip test that require local_machine
            item.add_marker(skip_not_local_machine)



import contextlib
import socket
import time
from multiprocessing import Event, Process

import pytest

from rpcclient.exceptions import RpcConnectionRefusedError, RpcFileNotFoundError, RpcResourceTemporarilyUnavailableError

RAND_PORT = 8989
BAD_SOCK = '/tmp/BAD'
TIMEOUT = 10
CHUNK_SIZE = 1024


def recvall(sock, size: int) -> bytes:
    buf = b''
    while len(buf) < size:
        buf += sock.recv(size - len(buf))
    return buf


def server_process_handler(server, event):
    event.set()
    client = server.accept()[0]
    while True:
        received = client.recv(CHUNK_SIZE)
        client.sendall(received)


@contextlib.contextmanager
def tcp_server():
    server = socket.socket()
    server.bind(('127.0.0.1', 0))
    port = server.getsockname()[1]
    server.listen(1)
    event = Event()
    server_process = Process(target=server_process_handler, args=(server, event,))
    server_process.start()
    event.wait()
    try:
        yield port
    finally:
        server_process.kill()


@contextlib.contextmanager
def unix_server(tmp_path):
    server = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock_path = tmp_path / 'sock'
    sock_path = str(sock_path.absolute())
    server.bind(sock_path)
    server.listen(1)
    event = Event()
    server_process = Process(target=server_process_handler, args=(server, event,))
    server_process.start()
    event.wait()
    try:
        yield sock_path
    finally:
        server_process.kill()


def test_tcp_connection_refused(client):
    with pytest.raises(RpcConnectionRefusedError):
        with client.network.tcp_connect('127.0.0.1', RAND_PORT):
            pass


def test_unix_connection_refused(client):
    with pytest.raises(RpcFileNotFoundError):
        with client.network.unix_connect(BAD_SOCK):
            pass


@pytest.mark.local_machine
@pytest.mark.parametrize('send_buf', [
    b'a',
    b'b' * 20,
    b'c' * 1024,
])
def test_tcp_send_receive(client, send_buf: bytes):
    with tcp_server() as port:
        with client.network.tcp_connect('127.0.0.1', port) as sock:
            sock.sendall(send_buf)
            assert recvall(sock, len(send_buf)) == send_buf


@pytest.mark.local_machine
@pytest.mark.parametrize('send_buf', [
    b'a',
    b'b' * 20,
    b'c' * 1024,
])
def test_unix_send_receive(client, tmp_path, send_buf: bytes):
    with unix_server(tmp_path) as sock_path:
        with client.network.unix_connect(sock_path) as sock:
            sock.sendall(send_buf)
            assert recvall(sock, len(send_buf)) == send_buf


@pytest.mark.local_machine
def test_tcp_receive_timeout(client):
    with tcp_server() as port:
        with client.network.tcp_connect('127.0.0.1', port) as sock:
            sock.settimeout(TIMEOUT)
            start = time.time()
            with pytest.raises(RpcResourceTemporarilyUnavailableError):
                sock.recv(1)
    assert time.time() - start >= TIMEOUT


@pytest.mark.local_machine
def test_unix_receive_timeout(client, tmp_path):
    with unix_server(tmp_path) as tmp_sock:
        with client.network.unix_connect(tmp_sock) as sock:
            sock.settimeout(TIMEOUT)
            start = time.time()
            with pytest.raises(RpcResourceTemporarilyUnavailableError):
                sock.recv(1)
        assert time.time() - start >= TIMEOUT



import typing
from collections import UserDict

import IPython
from pygments import formatters, highlight, lexers

from rpcclient.allocated import Allocated
from rpcclient.exceptions import RpcClientException

SHELL_USAGE = """
# Welcome to SCPreference plist interactive editor!
# Please use the `d` global in order to make changes to the current plist.
# Use: `d.commit()` to save your changes when done.
#
# For example, consider the following:

# view current plist
print(d)

# modify an item
d['item1'] = 5

# update
d.commit()

# That's it! hope you had a pleasant ride! 👋
"""


class SCPreference(Allocated):
    def __init__(self, client, preferences_id: str, ref):
        super().__init__()
        self._client = client
        self._ref = ref
        self._preferences_id = preferences_id

    @property
    def keys(self) -> typing.List[str]:
        """ wrapper for SCPreferencesCopyKeyList """
        return self._client.symbols.SCPreferencesCopyKeyList(self._ref).py()

    def _set(self, key: str, value):
        """ wrapper for SCPreferencesSetValue """
        if not self._client.symbols.SCPreferencesSetValue(self._ref, self._client.cf(key), self._client.cf(value)):
            raise RpcClientException(f'SCPreferencesSetValue failed to set: {key}')

    def set(self, key: str, value):
        """ set key:value and commit the change """
        self._set(key, value)
        self._commit()

    def set_dict(self, d: typing.Mapping):
        """ set the entire preference dictionary (clear if already exists) and commit the change """
        self._clear()
        self._update_dict(d)
        self._commit()

    def _update_dict(self, d: typing.Mapping):
        """ update preference dictionary """
        for k, v in d.items():
            self._set(k, v)

    def update_dict(self, d: typing.Mapping):
        """ update preference dictionary and commit """
        self._update_dict(d)
        self._commit()

    def _remove(self, key: str):
        """ wrapper for SCPreferencesRemoveValue """
        if not self._client.symbols.SCPreferencesRemoveValue(self._ref, self._client.cf(key)):
            raise RpcClientException(f'SCPreferencesRemoveValue failed to remove: {key}')

    def remove(self, key: str):
        """ remove given key and commit """
        self._remove(key)
        self._commit()

    def get(self, key: str):
        """ wrapper for SCPreferencesGetValue """
        return self._client.symbols.SCPreferencesGetValue(self._ref, self._client.cf(key)).py()

    def get_dict(self) -> typing.Mapping:
        """ get a dictionary representation """
        result = {}
        for k in self.keys:
            result[k] = self.get(k)
        return result

    def interactive(self):
        """ open an interactive IPython shell for viewing and editing """
        plist = Plist(self)
        IPython.embed(
            header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.TerminalTrueColorFormatter(style='native')),
            user_ns={
                'd': plist,
            })

    def _clear(self):
        """ clear dictionary """
        for k in self.keys:
            self._remove(k)

    def clear(self):
        """ clear dictionary and commit """
        self._clear()
        self._commit()

    def _deallocate(self):
        """ free the preference object """
        self._client.symbols.CFRelease(self._ref)

    def _commit(self):
        """ commit all changes """
        if not self._client.symbols.SCPreferencesCommitChanges(self._ref):
            raise RpcClientException('SCPreferencesCommitChanges failed')
        self._client.symbols.SCPreferencesSynchronize(self._ref)

    def __repr__(self):
        return f'<{self.__class__.__name__} NAME:{self._preferences_id}>'


class Plist(UserDict):
    def __init__(self, preference: SCPreference):
        super().__init__(preference.get_dict())
        self._preference = preference

    def commit(self):
        self._preference.set_dict(self)


class SCPreferences:
    """
    API to the SCPreferences* functions - preferences managed by SystemConfiguration framework.
    https://developer.apple.com/documentation/systemconfiguration/scpreferences?language=objc
    """

    def __init__(self, client):
        """
        :param rpcclient.darwin.client.DarwinClient client:
        """
        self._client = client

    def open(self, preferences_id: str) -> SCPreference:
        """ get an SCPreference from a given preferences_id """
        ref = self._client.symbols.SCPreferencesCreate(0, self._client.cf('rpcserver'), self._client.cf(preferences_id))
        if not ref:
            raise RpcClientException(f'SCPreferencesCreate failed for: {preferences_id}')
        return SCPreference(self._client, preferences_id, ref)

    def get_keys(self, preferences_id: str) -> typing.List[str]:
        """ get all keys from given preferences_id """
        with self.open(preferences_id) as o:
            return o.keys

    def get_dict(self, preferences_id: str):
        """ get dict from given preferences_id """
        with self.open(preferences_id) as o:
            return o.get_dict()



import dataclasses
import errno
import logging
import posixpath
import re
import struct
import time
from collections import namedtuple
from datetime import datetime
from pathlib import Path
from typing import Generator, List, Mapping, Optional

from cached_property import cached_property
from construct import Array, Container, Int32ul
from parameter_decorators import path_to_str

from rpcclient.darwin.consts import EXC_MASK_ALL, EXC_TYPES_COUNT, MACH_PORT_TYPE_ALL_RIGHTS, \
    MACH_PORT_TYPE_DEAD_NAME, MACH_PORT_TYPE_DNREQUEST, MACH_PORT_TYPE_PORT_SET, MACH_PORT_TYPE_RECEIVE, \
    MACH_PORT_TYPE_SEND, MACH_PORT_TYPE_SEND_ONCE, TASK_DYLD_INFO, TASK_FLAVOR_READ, THREAD_IDENTIFIER_INFO, \
    VM_FLAGS_ANYWHERE, ARMThreadFlavors, x86_THREAD_STATE64
from rpcclient.darwin.structs import ARM_THREAD_STATE64_COUNT, FAT_CIGAM, FAT_MAGIC, LOAD_COMMAND_TYPE, MAXPATHLEN, \
    PROC_PIDFDPIPEINFO, PROC_PIDFDSOCKETINFO, PROC_PIDFDVNODEPATHINFO, PROC_PIDLISTFDS, PROC_PIDTASKALLINFO, \
    PROX_FDTYPE_KQUEUE, PROX_FDTYPE_PIPE, PROX_FDTYPE_SOCKET, PROX_FDTYPE_VNODE, TASK_DYLD_INFO_COUNT, \
    THREAD_IDENTIFIER_INFO_COUNT, all_image_infos_t, arm_thread_state64_t, dyld_image_info_t, fat_header, \
    ipc_info_name_t, mach_header_t, mach_port_t, pid_t, pipe_info, proc_fdinfo, proc_taskallinfo, procargs2_t, \
    so_family_t, so_kind_t, socket_fdinfo, task_dyld_info_data_t, thread_identifier_info, vnode_fdinfowithpath, \
    x86_thread_state64_t
from rpcclient.darwin.symbol import DarwinSymbol
from rpcclient.exceptions import ArgumentError, BadReturnValueError, MissingLibraryError, ProcessSymbolAbsentError, \
    RpcClientException, SymbolAbsentError, UnrecognizedSelectorError
from rpcclient.processes import Processes
from rpcclient.protos.rpc_pb2 import ARCH_ARM64
from rpcclient.structs.consts import RTLD_NOW, SEEK_SET, SIGKILL, SIGTERM
from rpcclient.symbol import ADDRESS_SIZE_TO_STRUCT_FORMAT, Symbol
from rpcclient.sysctl import CTL, KERN

_CF_STRING_ARRAY_PREFIX_LEN = len('    "')
_CF_STRING_ARRAY_SUFFIX_LEN = len('",')
_BACKTRACE_FRAME_REGEX = re.compile(r'\[\s*(\d+)\] (0x[0-9a-f]+)\s+\{(.+?) \+ (.+?)\} (.*)')

FdStruct = namedtuple('FdStruct', 'fd struct')

logger = logging.getLogger(__name__)

CDHASH_SIZE = 20
CHUNK_SIZE = 1024 * 64
APP_SUFFIX = '.app/'


@dataclasses.dataclass()
class Fd:
    fd: int


@dataclasses.dataclass()
class KQueueFd(Fd):
    pass


@dataclasses.dataclass()
class PipeFd(Fd):
    pass


@dataclasses.dataclass()
class FileFd(Fd):
    path: str


@dataclasses.dataclass()
class UnixFd(Fd):
    path: str


@dataclasses.dataclass()
class SocketFd(Fd):
    pass


@dataclasses.dataclass()
class Ipv4SocketFd(SocketFd):
    local_address: str
    local_port: int
    remote_address: str
    remote_port: int  # when remote 0, the socket is for listening


@dataclasses.dataclass()
class Ipv6SocketFd(SocketFd):
    local_address: str
    local_port: int
    remote_address: str
    remote_port: int  # when remote 0, the socket is for listening


@dataclasses.dataclass()
class Ipv4TcpFd(Ipv4SocketFd):
    pass


@dataclasses.dataclass()
class Ipv6TcpFd(Ipv6SocketFd):
    pass


@dataclasses.dataclass()
class Ipv4UdpFd(Ipv4SocketFd):
    pass


@dataclasses.dataclass()
class Ipv6UdpFd(Ipv6SocketFd):
    pass


Image = namedtuple('Image', 'address path')

SOCKET_TYPE_DATACLASS = {
    so_family_t.AF_INET: {
        so_kind_t.SOCKINFO_TCP: Ipv4TcpFd,
        so_kind_t.SOCKINFO_IN: Ipv4UdpFd,
    },
    so_family_t.AF_INET6: {
        so_kind_t.SOCKINFO_TCP: Ipv6TcpFd,
        so_kind_t.SOCKINFO_IN: Ipv6UdpFd,
    }
}


class Thread:
    def __init__(self, client, thread_id: int):
        self._client = client
        self._thread_id = thread_id

    @property
    def thread_id(self) -> int:
        return self._thread_id

    def get_state(self):
        raise NotImplementedError()

    def set_state(self, state: Mapping):
        raise NotImplementedError()

    def resume(self):
        raise NotImplementedError()

    def suspend(self):
        raise NotImplementedError()

    def __repr__(self):
        return f'<{self.__class__.__name__} TID:{self._thread_id}>'


class IntelThread64(Thread):
    def get_state(self):
        with self._client.safe_malloc(x86_thread_state64_t.sizeof()) as p_state:
            with self._client.safe_malloc(x86_thread_state64_t.sizeof()) as p_thread_state_count:
                p_thread_state_count[0] = x86_thread_state64_t.sizeof() // Int32ul.sizeof()
                if self._client.symbols.thread_get_state(self._thread_id, x86_THREAD_STATE64,
                                                         p_state, p_thread_state_count):
                    raise BadReturnValueError('thread_get_state() failed')
                return x86_thread_state64_t.parse_stream(p_state)

    def set_state(self, state: Mapping):
        if self._client.symbols.thread_set_state(self._thread_id, x86_THREAD_STATE64,
                                                 x86_thread_state64_t.build(state),
                                                 x86_thread_state64_t.sizeof() // Int32ul.sizeof()):
            raise BadReturnValueError('thread_set_state() failed')


class ArmThread64(Thread):
    def get_state(self):
        with self._client.safe_malloc(arm_thread_state64_t.sizeof()) as p_state:
            with self._client.safe_malloc(arm_thread_state64_t.sizeof()) as p_thread_state_count:
                p_thread_state_count[0] = ARM_THREAD_STATE64_COUNT
                if self._client.symbols.thread_get_state(self._thread_id, ARMThreadFlavors.ARM_THREAD_STATE64,
                                                         p_state, p_thread_state_count):
                    raise BadReturnValueError('thread_get_state() failed')
                return arm_thread_state64_t.parse_stream(p_state)

    def set_state(self, state: Mapping):
        if self._client.symbols.thread_set_state(self._thread_id, ARMThreadFlavors.ARM_THREAD_STATE64,
                                                 arm_thread_state64_t.build(state),
                                                 ARM_THREAD_STATE64_COUNT):
            raise BadReturnValueError('thread_set_state() failed')

    def suspend(self):
        if self._client.symbols.thread_suspend(self._thread_id):
            raise BadReturnValueError('thread_suspend() failed')

    def resume(self):
        if self._client.symbols.thread_resume(self._thread_id):
            raise BadReturnValueError('thread_resume() failed')


@dataclasses.dataclass
class Region:
    region_type: str
    start: 'ProcessSymbol'
    end: int
    vsize: str
    protection: str
    protection_max: str
    region_detail: str

    @property
    def size(self) -> int:
        return self.end - self.start


@dataclasses.dataclass
class LoadedClass:
    name: str
    type_name: str
    binary_path: str


@dataclasses.dataclass
class Frame:
    depth: int
    address: int
    section: str
    offset: int
    symbol_name: str

    def __repr__(self):
        return f'<{self.__class__.__name__} [{self.depth:3}] 0x{self.address:x} ({self.section} + 0x{self.offset:x}) ' \
               f'{self.symbol_name}>'


@dataclasses.dataclass
class Backtrace:
    flavor: str
    time_start: float
    time_end: float
    pid: int
    thread_id: int
    dispatch_queue_serial_num: int
    frames: List[Frame]

    def __init__(self, vmu_backtrace: DarwinSymbol):
        backtrace = vmu_backtrace.objc_call('description').py()
        match = re.match(r'VMUBacktrace \(Flavor: (?P<flavor>.+?) Simple Time: (?P<time>.+?) '
                         r'Process: (?P<pid>\d+) Thread: (?P<thread_id>.+?)  Dispatch queue serial num: '
                         r'(?P<dispatch_queue_serial_num>\d+)\)', backtrace)
        self.flavor = match.group('flavor')
        self.pid = int(match.group('pid'))
        self.thread_id = int(match.group('thread_id'), 16)
        self.dispatch_queue_serial_num = int(match.group('dispatch_queue_serial_num'))

        self.frames = []
        for frame in re.findall(_BACKTRACE_FRAME_REGEX, backtrace):
            self.frames.append(Frame(depth=int(frame[0]), address=int(frame[1], 0), section=frame[2],
                                     offset=int(frame[3], 0), symbol_name=frame[4]))

    def __repr__(self):
        buf = f'<{self.__class__.__name__} PID: {self.pid} TID: {self.thread_id}\n'
        for frame in self.frames:
            buf += f'    {frame}\n'
        buf += '>'
        return buf


class ProcessSymbol(Symbol):
    @classmethod
    def create(cls, value: int, client, process):
        symbol = super().create(value, client)
        symbol = ProcessSymbol(symbol)
        symbol._prepare(client)
        symbol.process = process
        return symbol

    def _clone_from_value(self, value: int):
        return self.create(value, self._client, self.process)

    def peek(self, count: int) -> bytes:
        return self.process.peek(self, count)

    def poke(self, buf: bytes) -> None:
        return self.process.poke(self, buf)

    def peek_str(self, encoding='utf-8') -> str:
        """ peek string at given address """
        return self.process.peek_str(self, encoding)

    @property
    def dl_info(self) -> Container:
        raise NotImplementedError('dl_info isn\'t implemented for remote process symbols')

    @property
    def name(self) -> str:
        return self.process.get_symbol_name(self)

    @property
    def filename(self) -> str:
        return self.process.get_symbol_image(self.name).path

    def __getitem__(self, item) -> 'ProcessSymbol':
        fmt = ADDRESS_SIZE_TO_STRUCT_FORMAT[self.item_size]
        addr = self + item * self.item_size
        deref = struct.unpack(self._client._endianness + fmt, self.process.peek(addr, self.item_size))[0]
        return self.create(deref, self._client, self.process)

    def __setitem__(self, item, value):
        fmt = ADDRESS_SIZE_TO_STRUCT_FORMAT[self.item_size]
        value = struct.pack(self._client._endianness + fmt, int(value))
        self.process.poke(self + item * self.item_size, value)

    def __call__(self, *args, **kwargs):
        raise RpcClientException('ProcessSymbol is not callable')


@dataclasses.dataclass
class MachPortThreadInfo:
    thread_ids: List[int]


@dataclasses.dataclass
class MachPortInfo:
    task: int
    pid: int
    name: int
    rights: List[str]
    ipc_object: int
    dead: bool
    proc_name: Optional[str] = None
    thread_info: Optional[MachPortThreadInfo] = None

    @property
    def has_recv_right(self) -> bool:
        return 'recv' in self.rights

    @property
    def has_send_right(self) -> bool:
        return 'send' in self.rights


@dataclasses.dataclass
class MachPortCrossRefInfo:
    name: int
    ipc_object: int
    recv_right_pid: int
    recv_right_proc_name: str


class Process:
    PEEK_STR_CHUNK_SIZE = 0x100

    def __init__(self, client, pid: int):
        self._client = client
        self._pid = pid

        if self._client.arch == ARCH_ARM64:
            self._thread_class = ArmThread64
        else:
            self._thread_class = IntelThread64

    def kill(self, sig: int = SIGTERM):
        """ kill(pid, sig) at remote. read man for more details. """
        return self._client.processes.kill(self._pid, sig)

    def waitpid(self, flags: int = 0):
        """ waitpid(pid, stat_loc, 0) at remote. read man for more details. """
        return self._client.processes.waitpid(self._pid, flags)

    def peek(self, address: int, size: int) -> bytes:
        """ peek at memory address """
        with self._client.safe_malloc(size) as buf:
            with self._client.safe_malloc(8) as p_size:
                p_size[0] = size
                if self._client.symbols.vm_read_overwrite(self.task, address, size, buf, p_size):
                    raise BadReturnValueError('vm_read() failed')
                return buf.peek(size)

    def peek_str(self, address: int, encoding='utf-8') -> str:
        """ peek string at memory address """
        size = self.PEEK_STR_CHUNK_SIZE
        buf = b''

        while size:
            try:
                buf += self.peek(address, size)
                if b'\x00' in buf:
                    return buf.split(b'\x00', 1)[0].decode(encoding)
                address += size
            except BadReturnValueError:
                size = size // 2

    def poke(self, address: int, buf: bytes):
        """ poke at memory address """
        if self._client.symbols.vm_write(self.task, address, buf, len(buf)):
            raise BadReturnValueError('vm_write() failed')

    def get_symbol_name(self, address: int) -> str:
        if self._client.arch != ARCH_ARM64:
            raise NotImplementedError('implemented only on ARCH_ARM64')
        result = self.vmu_object_identifier.objc_call('symbolForAddress:', address, return_raw=True)
        if result.x0 == 0 and result.x1 == 0:
            raise SymbolAbsentError()
        return self._client.symbols.CSSymbolGetName(result.x0, result.x1).peek_str()

    def get_symbol_image(self, name: str) -> Image:
        for image in self.images:
            result = self.get_symbol_address(name, posixpath.basename(image.path))
            if result:
                return image

        raise ProcessSymbolAbsentError()

    def get_symbol_class_info(self, address: int) -> DarwinSymbol:
        return self.vmu_object_identifier.objc_call('classInfoForMemory:length:', address, 8)

    def get_symbol_address(self, name: str, lib: str = None) -> ProcessSymbol:
        if lib is not None:
            return ProcessSymbol.create(
                self.vmu_object_identifier.objc_call('addressOfSymbol:inLibrary:', name, lib).c_uint64, self._client,
                self)

        image = self.get_symbol_image(name)
        return self.get_symbol_address(name, posixpath.basename(image.path))

    @property
    def loaded_classes(self):
        realized_classes = self.vmu_object_identifier.objc_call('realizedClasses')

        for i in range(1, realized_classes.objc_call('count') + 1):
            class_info = realized_classes.objc_call('classInfoForIndex:', i)
            name = class_info.objc_call('className').py()
            type_name = class_info.objc_call('typeName').py()
            binary_path = class_info.objc_call('binaryPath').py()
            yield LoadedClass(name=name, type_name=type_name, binary_path=binary_path)

    @property
    def images(self) -> List[Image]:
        """ get loaded image list """
        result = []

        with self._client.safe_malloc(task_dyld_info_data_t.sizeof()) as dyld_info:
            with self._client.safe_calloc(8) as count:
                count[0] = TASK_DYLD_INFO_COUNT
                if self._client.symbols.task_info(self.task, TASK_DYLD_INFO, dyld_info, count):
                    raise BadReturnValueError('task_info(TASK_DYLD_INFO) failed')
                dyld_info_data = task_dyld_info_data_t.parse_stream(dyld_info)
        all_image_infos = all_image_infos_t.parse(
            self.peek(dyld_info_data.all_image_info_addr, dyld_info_data.all_image_info_size))

        buf = self.peek(all_image_infos.infoArray, all_image_infos.infoArrayCount * dyld_image_info_t.sizeof())
        for image in Array(all_image_infos.infoArrayCount, dyld_image_info_t).parse(buf):
            path = self.peek_str(image.imageFilePath)
            result.append(Image(address=self.get_process_symbol(image.imageLoadAddress), path=path))
        return result

    @property
    def app_images(self) -> List[Image]:
        return [image for image in self.images if APP_SUFFIX in image.path]

    @property
    def threads(self) -> List[Thread]:
        result = []
        with self._client.safe_malloc(8) as threads:
            with self._client.safe_malloc(4) as count:
                count.item_size = 4
                if self._client.symbols.task_threads(self.task, threads, count):
                    raise BadReturnValueError('task_threads() failed')

                for tid in Array(count[0].c_uint32, Int32ul).parse(threads[0].peek(count[0] * 4)):
                    result.append(self._thread_class(self._client, tid))
        return result

    @property
    def pid(self) -> int:
        """ get pid """
        return self._pid

    @property
    def fds(self) -> List[Fd]:
        """ get a list of process opened file descriptors """
        result = []
        for fdstruct in self.fd_structs:
            fd = fdstruct.fd
            parsed = fdstruct.struct

            if fd.proc_fdtype == PROX_FDTYPE_VNODE:
                result.append(FileFd(fd=fd.proc_fd, path=parsed.pvip.vip_path))

            elif fd.proc_fdtype == PROX_FDTYPE_KQUEUE:
                result.append(KQueueFd(fd=fd.proc_fd))

            elif fd.proc_fdtype == PROX_FDTYPE_PIPE:
                result.append(PipeFd(fd=fd.proc_fd))

            elif fd.proc_fdtype == PROX_FDTYPE_SOCKET:
                if parsed.psi.soi_kind in (so_kind_t.SOCKINFO_TCP, so_kind_t.SOCKINFO_IN):
                    correct_class = SOCKET_TYPE_DATACLASS[parsed.psi.soi_family][parsed.psi.soi_kind]

                    if parsed.psi.soi_kind == so_kind_t.SOCKINFO_TCP:
                        info = parsed.psi.soi_proto.pri_tcp.tcpsi_ini
                    else:
                        info = parsed.psi.soi_proto.pri_in
                    result.append(correct_class(fd=fd.proc_fd,
                                                local_address=info.insi_laddr.ina_46.i46a_addr4,
                                                local_port=info.insi_lport,
                                                remote_address=info.insi_faddr.ina_46.i46a_addr4,
                                                remote_port=info.insi_fport))

                elif parsed.psi.soi_kind == so_kind_t.SOCKINFO_UN:
                    result.append(UnixFd(fd=fd.proc_fd, path=parsed.psi.soi_proto.pri_un.unsi_addr.ua_sun.sun_path))

        return result

    @property
    def fd_structs(self) -> List[FdStruct]:
        """ get a list of process opened file descriptors as raw structs """
        result = []
        size = self._client.symbols.proc_pidinfo(self.pid, PROC_PIDLISTFDS, 0, 0, 0)

        vi_size = 8196  # should be enough for all structs
        with self._client.safe_malloc(vi_size) as vi_buf:
            with self._client.safe_malloc(size) as fdinfo_buf:
                size = int(self._client.symbols.proc_pidinfo(self.pid, PROC_PIDLISTFDS, 0, fdinfo_buf, size))
                if not size:
                    raise BadReturnValueError('proc_pidinfo(PROC_PIDLISTFDS) failed')

                for fd in Array(size // proc_fdinfo.sizeof(), proc_fdinfo).parse(fdinfo_buf.peek(size)):

                    if fd.proc_fdtype == PROX_FDTYPE_VNODE:
                        # file
                        vs = self._client.symbols.proc_pidfdinfo(self.pid, fd.proc_fd, PROC_PIDFDVNODEPATHINFO, vi_buf,
                                                                 vi_size)
                        if not vs:
                            if self._client.errno == errno.EBADF:
                                # lsof treats this as fine
                                continue
                            raise BadReturnValueError(
                                f'proc_pidinfo(PROC_PIDFDVNODEPATHINFO) failed for fd: {fd.proc_fd} '
                                f'({self._client.last_error})')

                        result.append(
                            FdStruct(fd=fd,
                                     struct=vnode_fdinfowithpath.parse(vi_buf.peek(vnode_fdinfowithpath.sizeof()))))

                    elif fd.proc_fdtype == PROX_FDTYPE_KQUEUE:
                        result.append(FdStruct(fd=fd, struct=None))

                    elif fd.proc_fdtype == PROX_FDTYPE_SOCKET:
                        # socket
                        vs = self._client.symbols.proc_pidfdinfo(self.pid, fd.proc_fd, PROC_PIDFDSOCKETINFO, vi_buf,
                                                                 vi_size)
                        if not vs:
                            if self._client.errno == errno.EBADF:
                                # lsof treats this as fine
                                continue
                            raise BadReturnValueError(
                                f'proc_pidinfo(PROC_PIDFDSOCKETINFO) failed ({self._client.last_error})')

                        result.append(FdStruct(fd=fd, struct=socket_fdinfo.parse(vi_buf.peek(vi_size))))

                    elif fd.proc_fdtype == PROX_FDTYPE_PIPE:
                        # pipe
                        vs = self._client.symbols.proc_pidfdinfo(self.pid, fd.proc_fd, PROC_PIDFDPIPEINFO, vi_buf,
                                                                 vi_size)
                        if not vs:
                            if self._client.errno == errno.EBADF:
                                # lsof treats this as fine
                                continue
                            raise BadReturnValueError(
                                f'proc_pidinfo(PROC_PIDFDPIPEINFO) failed ({self._client.last_error})')

                        result.append(
                            FdStruct(fd=fd,
                                     struct=pipe_info.parse(vi_buf.peek(pipe_info.sizeof()))))

            return result

    @property
    def task_all_info(self):
        """ get a list of process opened file descriptors """
        with self._client.safe_malloc(proc_taskallinfo.sizeof()) as pti:
            if not self._client.symbols.proc_pidinfo(self.pid, PROC_PIDTASKALLINFO, 0, pti, proc_taskallinfo.sizeof()):
                raise BadReturnValueError('proc_pidinfo(PROC_PIDTASKALLINFO) failed')
            return proc_taskallinfo.parse_stream(pti)

    @property
    def backtraces(self) -> List[Backtrace]:
        result = []
        backtraces = self._client.symbols.objc_getClass('VMUSampler').objc_call('sampleAllThreadsOfTask:', self.task)
        for i in range(backtraces.objc_call('count')):
            bt = backtraces.objc_call('objectAtIndex:', i)
            result.append(Backtrace(bt))
        return result

    @cached_property
    def vmu_proc_info(self) -> DarwinSymbol:
        return self._client.symbols.objc_getClass('VMUProcInfo').objc_call('alloc').objc_call('initWithTask:',
                                                                                              self.task)

    @cached_property
    def vmu_region_identifier(self) -> DarwinSymbol:
        return self._client.symbols.objc_getClass('VMUVMRegionIdentifier').objc_call('alloc').objc_call('initWithTask:',
                                                                                                        self.task)

    @cached_property
    def vmu_object_identifier(self) -> DarwinSymbol:
        return self._client.symbols.objc_getClass('VMUObjectIdentifier').objc_call('alloc').objc_call('initWithTask:',
                                                                                                      self.task)

    @cached_property
    def task(self) -> int:
        self_task_port = self._client.symbols.mach_task_self()

        if self.pid == self._client.pid:
            return self_task_port

        with self._client.safe_malloc(8) as p_task:
            if self._client.symbols.task_for_pid(self_task_port, self.pid, p_task):
                raise BadReturnValueError('task_for_pid() failed')
            return p_task[0].c_int64

    @cached_property
    def path(self) -> Optional[str]:
        """ call proc_pidpath(filename, ...) at remote. review xnu header for more details. """
        with self._client.safe_malloc(MAXPATHLEN) as path:
            path_len = self._client.symbols.proc_pidpath(self.pid, path, MAXPATHLEN)
            if not path_len:
                return None
            return path.peek(path_len).decode()

    @cached_property
    def basename(self) -> Optional[str]:
        path = self.path
        if not path:
            return None
        return Path(path).parts[-1]

    @cached_property
    def name(self) -> str:
        return self.task_all_info.pbsd.pbi_name

    @cached_property
    def ppid(self) -> int:
        return self.task_all_info.pbsd.pbi_ppid

    @cached_property
    def uid(self) -> int:
        return self.task_all_info.pbsd.pbi_uid

    @cached_property
    def gid(self) -> int:
        return self.task_all_info.pbsd.pbi_gid

    @cached_property
    def ruid(self) -> int:
        return self.task_all_info.pbsd.pbi_ruid

    @cached_property
    def rgid(self) -> int:
        return self.task_all_info.pbsd.pbi_rgid

    @cached_property
    def start_time(self) -> datetime:
        if self._client.arch != ARCH_ARM64:
            raise NotImplementedError('implemented only on ARCH_ARM64')
        val = self.vmu_proc_info.objc_call('startTime', return_raw=True)
        tv_sec = val.x0
        tv_nsec = val.x1
        return datetime.fromtimestamp(tv_sec + (tv_nsec / (10 ** 9)))

    @property
    def parent(self) -> 'Process':
        return Process(self._client, self.ppid)

    @property
    def environ(self) -> List[str]:
        return self.vmu_proc_info.objc_call('envVars').py()

    @property
    def arguments(self) -> List[str]:
        return self.vmu_proc_info.objc_call('arguments').py()

    @property
    def raw_procargs2(self) -> bytes:
        return self._client.sysctl.get(CTL.KERN, KERN.PROCARGS2, self.pid)

    @property
    def procargs2(self) -> Container:
        return procargs2_t.parse(self.raw_procargs2)

    @property
    def regions(self) -> List[Region]:
        result = []

        # remove the '()' wrapping the list:
        #   (
        #       "item1",
        #       "item2",
        #   )
        buf = self.vmu_region_identifier.objc_call('regions').cfdesc.split('\n')[1:-1]

        for line in buf:
            # remove line prefix and suffix and split into words
            line = line[_CF_STRING_ARRAY_PREFIX_LEN:-_CF_STRING_ARRAY_SUFFIX_LEN].split()
            i = 0

            region_type = line[i]
            i += 1

            while '-' not in line[i]:
                # skip till address range
                i += 1

            address_range = line[i]
            i += 1
            start, end = address_range.split('-')
            start = int(start, 16)
            end = int(end, 16)
            vsize = None
            if 'V=' in line[i]:
                vsize = line[i].split('V=', 1)[1].split(']', 1)[0]

            while '/' not in line[i]:
                i += 1

            protection = line[i].split('/')
            i += 1

            region_detail = None
            if len(line) >= i + 1:
                region_detail = line[i]

            result.append(Region(region_type=region_type, start=self.get_process_symbol(start), end=end, vsize=vsize,
                                 protection=protection[0], protection_max=protection[1], region_detail=region_detail))

        return result

    @property
    def cdhash(self) -> bytes:
        with self._client.safe_malloc(CDHASH_SIZE) as cdhash:
            # by reversing online-auth-agent
            if 0 != self._client.symbols.csops(self.pid, 5, cdhash, CDHASH_SIZE):
                raise BadReturnValueError(f'failed to get cdhash for {self.pid}')
            return cdhash.peek(CDHASH_SIZE)

    def get_process_symbol(self, address: int) -> ProcessSymbol:
        return ProcessSymbol.create(address, self._client, self)

    def vm_allocate(self, size: int) -> ProcessSymbol:
        with self._client.safe_malloc(8) as out_address:
            if self._client.symbols.vm_allocate(self.task, out_address, size, VM_FLAGS_ANYWHERE):
                raise BadReturnValueError('vm_allocate() failed')
            return self.get_process_symbol(out_address[0])

    @path_to_str('output_dir')
    def dump_app(self, output_dir: str, chunk_size=CHUNK_SIZE) -> None:
        """
        Based on:
        https://github.com/AloneMonkey/frida-ios-dump/blob/master/dump.js
        """
        for image in self.app_images:
            relative_name = image.path.split(APP_SUFFIX, 1)[1]
            output_file = Path(output_dir).expanduser() / relative_name
            output_file.parent.mkdir(exist_ok=True, parents=True)
            logger.debug(f'dumping: {output_file}')

            with open(output_file, 'wb') as output_file:
                macho_in_memory = mach_header_t.parse_stream(image.address)

                with self._client.fs.open(image.path, 'r') as fat_file:
                    # locating the correct MachO offset within the FAT image (if FAT)
                    magic_in_fs = Int32ul.parse_stream(fat_file)
                    fat_file.seek(0, SEEK_SET)
                    if magic_in_fs in (FAT_CIGAM, FAT_MAGIC):
                        parsed_fat = fat_header.parse_stream(fat_file)
                        for arch in parsed_fat.archs:
                            if arch.cputype == macho_in_memory.cputype and arch.cpusubtype == macho_in_memory.cpusubtype:
                                # correct MachO offset found
                                file_offset = arch.offset
                                break
                        fat_file.seek(file_offset, SEEK_SET)

                    # perform actual MachO dump
                    output_file.seek(0, SEEK_SET)
                    while True:
                        chunk = fat_file.read(chunk_size)
                        if len(chunk) == 0:
                            break
                        output_file.write(chunk)

                    offset_cryptid = None

                    # if image is encrypted, patch its encryption loader command
                    for load_command in macho_in_memory.load_commands:
                        if load_command.cmd in (LOAD_COMMAND_TYPE.LC_ENCRYPTION_INFO_64,
                                                LOAD_COMMAND_TYPE.LC_ENCRYPTION_INFO):
                            offset_cryptid = load_command.data.cryptid_offset - image.address
                            crypt_offset = load_command.data.cryptoff
                            crypt_size = load_command.data.cryptsize
                            break

                    if offset_cryptid is not None:
                        output_file.seek(offset_cryptid, SEEK_SET)
                        output_file.write(b'\x00' * 4)  # cryptid = 0
                        output_file.seek(crypt_offset, SEEK_SET)
                        output_file.flush()
                        output_file.write((image.address + crypt_offset).peek(crypt_size))

    def get_mach_port_cross_ref_info(self) -> List[MachPortCrossRefInfo]:
        """ Get all allocated mach ports and cross-refs to get the recv right owner """
        result = []
        own_ports = []
        cross_refs_info = {}
        for port in self._client.processes.get_mach_ports():
            if port.pid == self.pid:
                own_ports.append(port)
            if port.ipc_object not in cross_refs_info:
                cross_refs_info[port.ipc_object] = []
            cross_refs_info[port.ipc_object].append(port)

        for port in own_ports:
            for cross_ref_port_info in cross_refs_info[port.ipc_object]:
                if cross_ref_port_info.has_recv_right:
                    result.append(MachPortCrossRefInfo(name=port.name, ipc_object=port.ipc_object,
                                                       recv_right_pid=cross_ref_port_info.pid,
                                                       recv_right_proc_name=cross_ref_port_info.proc_name))
        return result

    def get_memgraph_snapshot(self) -> bytes:
        scanner = None
        try:
            # first attempt via new API
            scanner = self._client.symbols.objc_getClass('VMUProcessObjectGraph').objc_call(
                'createWithTask:', self.task)
            snapshot_graph = scanner.objc_call('plistRepresentationWithOptions:', 0).py()
        except UnrecognizedSelectorError:
            # if failed, attempt with old API
            scanner = self._client.symbols.objc_getClass('VMUTaskMemoryScanner').objc_call('alloc').objc_call(
                'initWithTask:', self.task)
            scanner.objc_call('addRootNodesFromTask')
            scanner.objc_call('addMallocNodesFromTask')
            snapshot_graph = scanner.objc_call('processSnapshotGraph').objc_call(
                'plistRepresentationWithOptions:', 0).py()
        finally:
            if scanner is not None:
                # free object scanner so process can resume
                scanner.objc_call('release')
        return snapshot_graph

    def __repr__(self):
        return f'<{self.__class__.__name__} PID:{self.pid} PATH:{self.path}>'


class DarwinProcesses(Processes):
    """ manage processes """

    def __init__(self, client):
        super().__init__(client)
        self._load_symbolication_library()
        self._self_process = Process(self._client, self._client.pid)

    def _load_symbolication_library(self):
        options = [
            '/System/Library/PrivateFrameworks/Symbolication.framework/Symbolication'
        ]
        for option in options:
            if self._client.dlopen(option, RTLD_NOW):
                return
        raise MissingLibraryError('Symbolication library isn\'t available')

    def get_self(self) -> Process:
        """ get self process """
        return self._self_process

    def get_by_pid(self, pid: int) -> Process:
        """ get process object by pid """
        proc_list = self.list()
        for p in proc_list:
            if p.pid == pid:
                return p
        raise ArgumentError(f'failed to locate process with pid: {pid}')

    def get_by_basename(self, name: str) -> Process:
        """ get process object by basename """
        proc_list = self.list()
        for p in proc_list:
            if p.basename == name:
                return p
        raise ArgumentError(f'failed to locate process with name: {name}')

    def get_by_name(self, name: str) -> Process:
        """ get process object by name """
        proc_list = self.list()
        for p in proc_list:
            if p.name == name:
                return p
        raise ArgumentError(f'failed to locate process with name: {name}')

    def grep(self, name: str) -> List[Process]:
        """ get process list by basename filter """
        result = []
        proc_list = self.list()
        for p in proc_list:
            if p.basename and name in p.basename:
                result.append(p)
        return result

    def get_processes_by_listening_port(self, port: int) -> List[Process]:
        """ get a process object listening on the specified port """
        listening_processes = []
        for process in self.list():
            try:
                fds = process.fds
            except BadReturnValueError:
                # it's possible to get error if new processes have since died or the rpcserver
                # doesn't have the required permissions to access all the processes
                continue

            for fd in fds:
                if (isinstance(fd, Ipv4SocketFd) or isinstance(fd, Ipv6SocketFd)) and \
                        fd.local_port == port and fd.remote_port == 0:
                    listening_processes.append(process)
        return listening_processes

    def lsof(self) -> Mapping[int, List[Fd]]:
        """ get dictionary of pid to its opened fds """
        result = {}
        for process in self.list():
            try:
                fds = process.fds
            except BadReturnValueError:
                # it's possible to get error if new processes have since died or the rpcserver
                # doesn't have the required permissions to access all the processes
                continue

            result[process.pid] = fds
        return result

    @path_to_str('path')
    def fuser(self, path: str) -> List[Process]:
        """get a list of all processes have an open hande to the specified path """
        result = []
        proc_list = self.list()
        for process in proc_list:
            try:
                fds = process.fds
            except BadReturnValueError:
                # it's possible to get error if new processes have since died or the rpcserver
                # doesn't have the required permissions to access all the processes
                continue

            for fd in fds:
                if isinstance(fd, FileFd):
                    if str(Path(fd.path).absolute()) == str(Path(path).absolute()):
                        result.append(process)

        return result

    def list(self) -> List[Process]:
        """ list all currently running processes """
        n = self._client.symbols.proc_listallpids(0, 0)
        pid_buf_size = pid_t.sizeof() * n
        with self._client.safe_malloc(pid_buf_size) as pid_buf:
            pid_buf.item_size = pid_t.sizeof()
            n = self._client.symbols.proc_listallpids(pid_buf, pid_buf_size)

            result = []
            for i in range(n):
                pid = int(pid_buf[i])
                result.append(Process(self._client, pid))
            return result

    def disable_watchdog(self) -> None:
        while True:
            try:
                self.get_by_basename('watchdogd').kill(SIGKILL)
            except ArgumentError:
                pass
            time.sleep(1)

    def get_mach_ports(self, include_thread_info: bool = False) -> Generator[MachPortInfo, None, None]:  # noqa: C901
        """ Enumerate all mach ports (heavily inspired by lsmp) """
        thread_info = None
        mach_task_self = self._client.symbols.mach_task_self()

        p_psets = self._client.symbols.malloc(8)
        p_pset_count = self._client.symbols.malloc(8)
        p_pset_priv = self._client.symbols.malloc(8)
        p_tasks = self._client.symbols.malloc(8)
        p_task_count = self._client.symbols.malloc(8)
        p_count = self._client.symbols.malloc(4)
        ports_info = self._client.symbols.malloc(4 * 2 * EXC_TYPES_COUNT)
        masks = self._client.symbols.malloc(4 * EXC_TYPES_COUNT)
        behaviors = self._client.symbols.malloc(4 * EXC_TYPES_COUNT)
        flavors = self._client.symbols.malloc(4 * EXC_TYPES_COUNT)
        p_thread_count = self._client.symbols.malloc(4)
        p_thread_ports = self._client.symbols.malloc(8)
        info = self._client.symbols.malloc(200)
        p_pid = self._client.symbols.malloc(4)
        th_info = self._client.symbols.malloc(thread_identifier_info.sizeof())
        p_th_kobject = self._client.symbols.malloc(4)
        p_th_kotype = self._client.symbols.malloc(4)
        p_th_info_count = self._client.symbols.malloc(4)
        p_th_voucher = self._client.symbols.calloc(4, 1)
        p_table = self._client.symbols.malloc(8)
        unused = self._client.symbols.malloc(200)
        proc_name = self._client.symbols.malloc(100)
        p_kotype = self._client.symbols.malloc(4)

        p_pid.item_size = 4
        p_count.item_size = 4
        p_thread_count.item_size = 4
        p_th_info_count.item_size = 4
        p_th_voucher.item_size = 4
        p_th_info_count[0] = THREAD_IDENTIFIER_INFO_COUNT
        p_kotype.item_size = 4

        if self._client.symbols.getuid() == 0:
            # if privileged, get the info for all tasks so we can match ports up
            if self._client.symbols.host_processor_sets(self._client.symbols.mach_host_self(), p_psets,
                                                        p_pset_count) != 0:
                raise BadReturnValueError('host_processor_sets() failed')
            if p_pset_count[0] != 1:
                raise BadReturnValueError('Assertion Failure: pset count greater than one')

            # convert the processor-set-name port to a privileged port
            if self._client.symbols.host_processor_set_priv(self._client.symbols.mach_host_self(), p_psets[0][0],
                                                            p_pset_priv) != 0:
                raise BadReturnValueError('host_processor_set_priv() failed')

            self._client.symbols.mach_port_deallocate(mach_task_self, p_psets[0][0])
            self._client.symbols.vm_deallocate(mach_task_self, p_psets[0], p_pset_count[0] * mach_port_t.sizeof())

            # convert the processor-set-priv to a list of task read ports for the processor set
            if self._client.symbols.processor_set_tasks_with_flavor(p_pset_priv[0], TASK_FLAVOR_READ, p_tasks,
                                                                    p_task_count) != 0:
                raise BadReturnValueError('processor_set_tasks_with_flavor() failed')

            self._client.symbols.vm_deallocate(mach_task_self, p_pset_priv[0])

            # swap my current instances port to be last to collect all threads and exception port info
            my_task_position = None
            task_count = p_task_count[0]
            tasks = list(struct.unpack(f'<{int(task_count)}I', p_tasks[0].peek(task_count * 4)))

            for i in range(task_count):
                if self._client.symbols.mach_task_is_self(tasks[i]):
                    my_task_position = i
                    break
            if my_task_position is not None:
                swap_holder = tasks[task_count - 1]
                tasks[task_count - 1] = tasks[my_task_position]
                tasks[my_task_position] = swap_holder
        else:
            logger.warning('should run as root for best output (cross-ref to other tasks\' ports)')
            # just the one process
            task_count = 1
            with self._client.safe_malloc(8) as p_task:
                ret = self._client.symbols.task_read_for_pid(mach_task_self, self.pid, p_task)
                if ret != 0:
                    raise BadReturnValueError('task_read_for_pid() failed')
                tasks = [p_task[0]]

        for task in tasks:
            self._client.symbols.pid_for_task(task, p_pid)
            pid = p_pid[0].c_uint32

            if self._client.symbols.task_get_exception_ports_info(task, EXC_MASK_ALL, masks, p_count, ports_info,
                                                                  behaviors, flavors) != 0:
                raise BadReturnValueError('task_get_exception_ports_info() failed')

            if include_thread_info:
                # collect threads port as well
                if 0 != self._client.symbols.task_threads(task, p_thread_ports, p_thread_count):
                    raise BadReturnValueError('task_threads() failed')

                # collect the thread information
                thread_count = p_thread_count[0].c_uint32
                thread_ports = struct.unpack(f'<{thread_count}I', p_thread_ports[0].peek(4 * thread_count))
                thread_ids = []
                for thread_port in thread_ports:
                    ret = self._client.symbols.thread_get_exception_ports_info(
                        thread_port, EXC_MASK_ALL, masks, p_count, ports_info, behaviors, flavors)
                    if 0 != ret:
                        raise BadReturnValueError(
                            f'thread_get_exception_ports_info() failed: '
                            f'{self._client.symbols.mach_error_string(ret).peek_str()}')

                    if 0 == self._client.symbols.mach_port_kernel_object(mach_task_self, thread_port, p_th_kotype,
                                                                         p_th_kobject):
                        if 0 == self._client.symbols.thread_info(mach_task_self, thread_port,
                                                                 THREAD_IDENTIFIER_INFO, th_info, p_th_info_count[0]):
                            thread_id = thread_identifier_info.parse(
                                th_info.peek(thread_identifier_info.sizeof())).thread_id
                            thread_ids.append(thread_id)

                    self._client.symbols.mach_port_deallocate(mach_task_self, thread_port)
                thread_info = MachPortThreadInfo(thread_ids=thread_ids)

            if 0 != self._client.symbols.mach_port_space_info(task, info, p_table, p_count, unused, unused):
                raise BadReturnValueError('mach_port_space_info() failed')

            if 0 != self._client.symbols.proc_name(pid, proc_name, 100):
                proc_name_str = proc_name.peek_str()
            else:
                proc_name_str = None

            count = p_count[0].c_uint32
            table_struct = Array(count, ipc_info_name_t)

            parsed_table = table_struct.parse(p_table[0].peek(table_struct.sizeof()))

            for entry in parsed_table:
                dnreq = False
                rights = []

                if entry.iin_type & MACH_PORT_TYPE_ALL_RIGHTS == 0:
                    # skip empty slots in the table
                    continue

                if entry.iin_type == MACH_PORT_TYPE_PORT_SET:
                    continue

                if entry.iin_type & MACH_PORT_TYPE_SEND:
                    rights.append('send')

                if entry.iin_type & MACH_PORT_TYPE_DNREQUEST:
                    dnreq = True

                if entry.iin_type & MACH_PORT_TYPE_RECEIVE:
                    rights.append('recv')

                elif entry.iin_type == MACH_PORT_TYPE_DEAD_NAME:
                    continue

                if entry.iin_type == MACH_PORT_TYPE_SEND_ONCE:
                    pass

                yield MachPortInfo(
                    task=task, pid=pid, name=entry.iin_name, rights=rights, ipc_object=entry.iin_object, dead=dnreq,
                    proc_name=proc_name_str, thread_info=thread_info)



from construct import Bytes, Computed, Int8ul, Int16ul, Int32ul, Int64ul, PaddedString, Padding, Struct

_UTSNAME_LENGTH = 65
_D_NAME_LENGTH = 256

utsname = Struct(
    'sysname' / PaddedString(_UTSNAME_LENGTH, 'utf8'),
    'nodename' / PaddedString(_UTSNAME_LENGTH, 'utf8'),
    'release' / PaddedString(_UTSNAME_LENGTH, 'utf8'),
    'version' / PaddedString(_UTSNAME_LENGTH, 'utf8'),
    'machine' / PaddedString(_UTSNAME_LENGTH, 'utf8'),
    'domainname' / PaddedString(_UTSNAME_LENGTH, 'utf8'),
)

dirent = Struct(
    'd_ino' / Int32ul,
    Padding(4),
    'd_off' / Int64ul,
    'd_reclen' / Int16ul,
    'd_type' / Int8ul,
    '_d_name_bytes' / Bytes(_D_NAME_LENGTH),
    'd_name' / Computed(lambda x: x._d_name_bytes.split(b'\x00', 1)[0].decode())
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rpc.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\trpc.proto\x12\x03rpc\"\xb9\x04\n\x07\x43ommand\x12\r\n\x05magic\x18\x01 \x01(\x05\x12\x1c\n\x04\x65xec\x18\x02 \x01(\x0b\x32\x0c.rpc.CmdExecH\x00\x12 \n\x06\x64lopen\x18\x03 \x01(\x0b\x32\x0e.rpc.CmdDlopenH\x00\x12\"\n\x07\x64lclose\x18\x04 \x01(\x0b\x32\x0f.rpc.CmdDlcloseH\x00\x12\x1e\n\x05\x64lsym\x18\x05 \x01(\x0b\x32\r.rpc.CmdDlsymH\x00\x12\x1c\n\x04\x63\x61ll\x18\x06 \x01(\x0b\x32\x0c.rpc.CmdCallH\x00\x12\x1c\n\x04peek\x18\x07 \x01(\x0b\x32\x0c.rpc.CmdPeekH\x00\x12\x1c\n\x04poke\x18\x08 \x01(\x0b\x32\x0c.rpc.CmdPokeH\x00\x12#\n\x08list_dir\x18\t \x01(\x0b\x32\x0f.rpc.CmdListDirH\x00\x12)\n\x0bshow_object\x18\n \x01(\x0b\x32\x12.rpc.CmdShowObjectH\x00\x12\'\n\nshow_class\x18\x0b \x01(\x0b\x32\x11.rpc.CmdShowClassH\x00\x12)\n\x0b\x64ummy_block\x18\x0c \x01(\x0b\x32\x12.rpc.CmdDummyBlockH\x00\x12\x1e\n\x05\x63lose\x18\r \x01(\x0b\x32\r.rpc.CmdCloseH\x00\x12*\n\nclass_list\x18\x0e \x01(\x0b\x32\x14.rpc.CmdGetClassListH\x00\x12\'\n\nexec_chunk\x18\x0f \x01(\x0b\x32\x11.rpc.CmdExecChunkH\x00\x12 \n\x06\x63ustom\x18\x10 \x01(\x0b\x32\x0e.rpc.CmdCustomH\x00\x42\x06\n\x04type\"\xfc\x04\n\x08Response\x12$\n\x04\x65xec\x18\x01 \x01(\x0b\x32\x14.rpc.ResponseCmdExecH\x00\x12/\n\nexec_chunk\x18\x02 \x01(\x0b\x32\x19.rpc.ResponseCmdExecChunkH\x00\x12%\n\x06\x64lopen\x18\x03 \x01(\x0b\x32\x13.rpc.ResponseDlopenH\x00\x12\'\n\x07\x64lclose\x18\x04 \x01(\x0b\x32\x14.rpc.ResponseDlcloseH\x00\x12#\n\x05\x64lsym\x18\x05 \x01(\x0b\x32\x12.rpc.ResponseDlsymH\x00\x12!\n\x04peek\x18\x06 \x01(\x0b\x32\x11.rpc.ResponsePeekH\x00\x12!\n\x04poke\x18\x07 \x01(\x0b\x32\x11.rpc.ResponsePokeH\x00\x12!\n\x04\x63\x61ll\x18\x08 \x01(\x0b\x32\x11.rpc.ResponseCallH\x00\x12#\n\x05\x65rror\x18\t \x01(\x0b\x32\x12.rpc.ResponseErrorH\x00\x12.\n\x0b\x64ummy_block\x18\n \x01(\x0b\x32\x17.rpc.ResponseDummyBlockH\x00\x12.\n\x0bshow_object\x18\x0b \x01(\x0b\x32\x17.rpc.ResponseShowObjectH\x00\x12/\n\nclass_list\x18\x0c \x01(\x0b\x32\x19.rpc.ResponseGetClassListH\x00\x12,\n\nshow_class\x18\r \x01(\x0b\x32\x16.rpc.ResponseShowClassH\x00\x12(\n\x08list_dir\x18\x0e \x01(\x0b\x32\x14.rpc.ResponseListdirH\x00\x12%\n\x06\x63ustom\x18\x0f \x01(\x0b\x32\x13.rpc.ResponseCustomH\x00\x42\x06\n\x04type\"\xd4\x01\n\x12ReturnRegistersArm\x12\n\n\x02x0\x18\x01 \x01(\x04\x12\n\n\x02x1\x18\x02 \x01(\x04\x12\n\n\x02x2\x18\x03 \x01(\x04\x12\n\n\x02x3\x18\x04 \x01(\x04\x12\n\n\x02x4\x18\x05 \x01(\x04\x12\n\n\x02x5\x18\x06 \x01(\x04\x12\n\n\x02x6\x18\x07 \x01(\x04\x12\n\n\x02x7\x18\x08 \x01(\x04\x12\n\n\x02\x64\x30\x18\t \x01(\x01\x12\n\n\x02\x64\x31\x18\n \x01(\x01\x12\n\n\x02\x64\x32\x18\x0b \x01(\x01\x12\n\n\x02\x64\x33\x18\x0c \x01(\x01\x12\n\n\x02\x64\x34\x18\r \x01(\x01\x12\n\n\x02\x64\x35\x18\x0e \x01(\x01\x12\n\n\x02\x64\x36\x18\x0f \x01(\x01\x12\n\n\x02\x64\x37\x18\x10 \x01(\x01\"[\n\x08\x41rgument\x12\x0f\n\x05v_int\x18\x01 \x01(\x04H\x00\x12\x12\n\x08v_double\x18\x02 \x01(\x01H\x00\x12\x0f\n\x05v_str\x18\x03 \x01(\tH\x00\x12\x11\n\x07v_bytes\x18\x04 \x01(\x0cH\x00\x42\x06\n\x04type\"U\n\tHandshake\x12\r\n\x05magic\x18\x01 \x01(\r\x12\x17\n\x04\x61rch\x18\x02 \x01(\x0e\x32\t.rpc.Arch\x12\x0f\n\x07sysname\x18\x03 \x01(\t\x12\x0f\n\x07machine\x18\x04 \x01(\t\"*\n\tObjcClass\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x01(\t\" \n\rCmdShowObject\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x04\")\n\x12ResponseShowObject\x12\x13\n\x0b\x64\x65scription\x18\x01 \x01(\t\"\x1f\n\x0c\x43mdShowClass\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x04\"(\n\x11ResponseShowClass\x12\x13\n\x0b\x64\x65scription\x18\x01 \x01(\t\"\x1e\n\x0c\x43mdExecChunk\x12\x0e\n\x06\x62uffer\x18\x01 \x01(\x0c\"E\n\x14ResponseCmdExecChunk\x12\x10\n\x06\x62uffer\x18\x01 \x01(\x0cH\x00\x12\x13\n\texit_code\x18\x02 \x01(\rH\x00\x42\x06\n\x04type\"+\n\tCmdDlopen\x12\x10\n\x08\x66ilename\x18\x01 \x01(\t\x12\x0c\n\x04mode\x18\x02 \x01(\x05\" \n\x0eResponseDlopen\x12\x0e\n\x06handle\x18\x01 \x01(\x04\"\x1c\n\nCmdDlclose\x12\x0e\n\x06handle\x18\x01 \x01(\x04\"\x1e\n\x0fResponseDlclose\x12\x0b\n\x03res\x18\x01 \x01(\r\"/\n\x08\x43mdDlsym\x12\x0e\n\x06handle\x18\x01 \x01(\x04\x12\x13\n\x0bsymbol_name\x18\x02 \x01(\t\"\x1c\n\rResponseDlsym\x12\x0b\n\x03ptr\x18\x01 \x01(\x04\"9\n\x07\x43mdExec\x12\x12\n\nbackground\x18\x01 \x01(\x08\x12\x0c\n\x04\x61rgv\x18\x02 \x03(\t\x12\x0c\n\x04\x65nvp\x18\x03 \x03(\t\"\x1e\n\x0fResponseCmdExec\x12\x0b\n\x03pid\x18\x01 \x01(\r\"N\n\x07\x43mdCall\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x04\x12\x15\n\rva_list_index\x18\x02 \x01(\x04\x12\x1b\n\x04\x61rgv\x18\x03 \x03(\x0b\x32\r.rpc.Argument\"i\n\x0cResponseCall\x12\x30\n\rarm_registers\x18\x01 \x01(\x0b\x32\x17.rpc.ReturnRegistersArmH\x00\x12\x16\n\x0creturn_value\x18\x02 \x01(\x04H\x00\x42\x0f\n\rreturn_values\"(\n\x07\x43mdPeek\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x04\x12\x0c\n\x04size\x18\x02 \x01(\x04\"\x1c\n\x0cResponsePeek\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"(\n\x07\x43mdPoke\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x04\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\"\x0e\n\x0cResponsePoke\"\x1a\n\nCmdListDir\x12\x0c\n\x04path\x18\x01 \x01(\t\"\x0f\n\rCmdDummyBlock\"3\n\x12ResponseDummyBlock\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x04\x12\x0c\n\x04size\x18\x02 \x01(\x04\"\x11\n\x0f\x43mdGetClassList\"7\n\x14ResponseGetClassList\x12\x1f\n\x07\x63lasses\x18\x01 \x03(\x0b\x32\x0e.rpc.ObjcClass\"\x0f\n\rResponseError\"R\n\x0fResponseListdir\x12\r\n\x05magic\x18\x01 \x01(\x04\x12\x0c\n\x04\x64irp\x18\x02 \x01(\x04\x12\"\n\x0b\x64ir_entries\x18\x03 \x03(\x0b\x32\r.rpc.DirEntry\"m\n\x08\x44irEntry\x12\x0e\n\x06\x64_type\x18\x01 \x01(\x04\x12\x0e\n\x06\x64_name\x18\x02 \x01(\t\x12 \n\x05lstat\x18\x03 \x01(\x0b\x32\x11.rpc.DirEntryStat\x12\x1f\n\x04stat\x18\x04 \x01(\x0b\x32\x11.rpc.DirEntryStat\"\x83\x02\n\x0c\x44irEntryStat\x12\x0e\n\x06\x65rrno1\x18\x01 \x01(\x04\x12\x0e\n\x06st_dev\x18\x02 \x01(\x04\x12\x0f\n\x07st_mode\x18\x03 \x01(\x04\x12\x10\n\x08st_nlink\x18\x04 \x01(\x04\x12\x0e\n\x06st_ino\x18\x05 \x01(\x04\x12\x0e\n\x06st_uid\x18\x06 \x01(\x04\x12\x0e\n\x06st_gid\x18\x07 \x01(\x04\x12\x0f\n\x07st_rdev\x18\x08 \x01(\x04\x12\x0f\n\x07st_size\x18\t \x01(\x04\x12\x11\n\tst_blocks\x18\n \x01(\x04\x12\x12\n\nst_blksize\x18\x0b \x01(\x04\x12\x11\n\tst_atime1\x18\x0c \x01(\x04\x12\x11\n\tst_mtime1\x18\r \x01(\x04\x12\x11\n\tst_ctime1\x18\x0e \x01(\x04\"\n\n\x08\x43mdClose\"\x1c\n\tCmdCustom\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\"!\n\x0eResponseCustom\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c*(\n\x04\x41rch\x12\x10\n\x0c\x41RCH_UNKNOWN\x10\x00\x12\x0e\n\nARCH_ARM64\x10\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'rpc_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_ARCH']._serialized_start=3276
  _globals['_ARCH']._serialized_end=3316
  _globals['_COMMAND']._serialized_start=19
  _globals['_COMMAND']._serialized_end=588
  _globals['_RESPONSE']._serialized_start=591
  _globals['_RESPONSE']._serialized_end=1227
  _globals['_RETURNREGISTERSARM']._serialized_start=1230
  _globals['_RETURNREGISTERSARM']._serialized_end=1442
  _globals['_ARGUMENT']._serialized_start=1444
  _globals['_ARGUMENT']._serialized_end=1535
  _globals['_HANDSHAKE']._serialized_start=1537
  _globals['_HANDSHAKE']._serialized_end=1622
  _globals['_OBJCCLASS']._serialized_start=1624
  _globals['_OBJCCLASS']._serialized_end=1666
  _globals['_CMDSHOWOBJECT']._serialized_start=1668
  _globals['_CMDSHOWOBJECT']._serialized_end=1700
  _globals['_RESPONSESHOWOBJECT']._serialized_start=1702
  _globals['_RESPONSESHOWOBJECT']._serialized_end=1743
  _globals['_CMDSHOWCLASS']._serialized_start=1745
  _globals['_CMDSHOWCLASS']._serialized_end=1776
  _globals['_RESPONSESHOWCLASS']._serialized_start=1778
  _globals['_RESPONSESHOWCLASS']._serialized_end=1818
  _globals['_CMDEXECCHUNK']._serialized_start=1820
  _globals['_CMDEXECCHUNK']._serialized_end=1850
  _globals['_RESPONSECMDEXECCHUNK']._serialized_start=1852
  _globals['_RESPONSECMDEXECCHUNK']._serialized_end=1921
  _globals['_CMDDLOPEN']._serialized_start=1923
  _globals['_CMDDLOPEN']._serialized_end=1966
  _globals['_RESPONSEDLOPEN']._serialized_start=1968
  _globals['_RESPONSEDLOPEN']._serialized_end=2000
  _globals['_CMDDLCLOSE']._serialized_start=2002
  _globals['_CMDDLCLOSE']._serialized_end=2030
  _globals['_RESPONSEDLCLOSE']._serialized_start=2032
  _globals['_RESPONSEDLCLOSE']._serialized_end=2062
  _globals['_CMDDLSYM']._serialized_start=2064
  _globals['_CMDDLSYM']._serialized_end=2111
  _globals['_RESPONSEDLSYM']._serialized_start=2113
  _globals['_RESPONSEDLSYM']._serialized_end=2141
  _globals['_CMDEXEC']._serialized_start=2143
  _globals['_CMDEXEC']._serialized_end=2200
  _globals['_RESPONSECMDEXEC']._serialized_start=2202
  _globals['_RESPONSECMDEXEC']._serialized_end=2232
  _globals['_CMDCALL']._serialized_start=2234
  _globals['_CMDCALL']._serialized_end=2312
  _globals['_RESPONSECALL']._serialized_start=2314
  _globals['_RESPONSECALL']._serialized_end=2419
  _globals['_CMDPEEK']._serialized_start=2421
  _globals['_CMDPEEK']._serialized_end=2461
  _globals['_RESPONSEPEEK']._serialized_start=2463
  _globals['_RESPONSEPEEK']._serialized_end=2491
  _globals['_CMDPOKE']._serialized_start=2493
  _globals['_CMDPOKE']._serialized_end=2533
  _globals['_RESPONSEPOKE']._serialized_start=2535
  _globals['_RESPONSEPOKE']._serialized_end=2549
  _globals['_CMDLISTDIR']._serialized_start=2551
  _globals['_CMDLISTDIR']._serialized_end=2577
  _globals['_CMDDUMMYBLOCK']._serialized_start=2579
  _globals['_CMDDUMMYBLOCK']._serialized_end=2594
  _globals['_RESPONSEDUMMYBLOCK']._serialized_start=2596
  _globals['_RESPONSEDUMMYBLOCK']._serialized_end=2647
  _globals['_CMDGETCLASSLIST']._serialized_start=2649
  _globals['_CMDGETCLASSLIST']._serialized_end=2666
  _globals['_RESPONSEGETCLASSLIST']._serialized_start=2668
  _globals['_RESPONSEGETCLASSLIST']._serialized_end=2723
  _globals['_RESPONSEERROR']._serialized_start=2725
  _globals['_RESPONSEERROR']._serialized_end=2740
  _globals['_RESPONSELISTDIR']._serialized_start=2742
  _globals['_RESPONSELISTDIR']._serialized_end=2824
  _globals['_DIRENTRY']._serialized_start=2826
  _globals['_DIRENTRY']._serialized_end=2935
  _globals['_DIRENTRYSTAT']._serialized_start=2938
  _globals['_DIRENTRYSTAT']._serialized_end=3197
  _globals['_CMDCLOSE']._serialized_start=3199
  _globals['_CMDCLOSE']._serialized_end=3209
  _globals['_CMDCUSTOM']._serialized_start=3211
  _globals['_CMDCUSTOM']._serialized_end=3239
  _globals['_RESPONSECUSTOM']._serialized_start=3241
  _globals['_RESPONSECUSTOM']._serialized_end=3274
# @@protoc_insertion_point(module_scope)



from rpcclient.darwin.cfpreferences import CFPreferences
from rpcclient.darwin.scpreferences import SCPreferences


class Preferences:
    """ Preferences utils """

    def __init__(self, client):
        """
        :param rpcclient.darwin.client.DarwinClient client:
        """
        self.cf = CFPreferences(client)
        self.sc = SCPreferences(client)



import logging
from typing import List, Mapping

from rpcclient.exceptions import BadReturnValueError, RpcPermissionError
from rpcclient.symbol import Symbol

logger = logging.getLogger(__name__)


class Keychain:
    """ keychain utils """

    def __init__(self, client):
        self._client = client

    def add_internet_password(self, account: str, server: str, password: str):
        attributes = self._client.symbols.objc_getClass('NSMutableDictionary').objc_call('new')
        attributes.objc_call('setObject:forKey:', self._client.symbols.kSecClassInternetPassword[0],
                             self._client.symbols.kSecClass[0])
        attributes.objc_call('setObject:forKey:', self._client.cf(account), self._client.symbols.kSecAttrAccount[0])
        attributes.objc_call('setObject:forKey:', self._client.cf(server), self._client.symbols.kSecAttrServer[0])
        attributes.objc_call('setObject:forKey:', self._client.cf(password), self._client.symbols.kSecValueData[0])
        err = self._client.symbols.SecItemAdd(attributes, 0).c_int32
        if err != 0:
            raise BadReturnValueError(f'SecItemAdd() returned: {err}')

    def query_apple_share_passwords(self) -> List[Mapping]:
        return self._query(self._client.symbols.kSecClassAppleSharePassword)

    def query_internet_passwords(self) -> List[Mapping]:
        return self._query(self._client.symbols.kSecClassInternetPassword)

    def query_generic_passwords(self) -> List[Mapping]:
        return self._query(self._client.symbols.kSecClassGenericPassword)

    def query_identities(self) -> List[Mapping]:
        return self._query(self._client.symbols.kSecClassIdentity)

    def query_certificates(self) -> List[Mapping]:
        return self._query(self._client.symbols.kSecClassCertificate)

    def query_keys(self) -> List[Mapping]:
        return self._query(self._client.symbols.kSecClassKey)

    def _query(self, class_type: Symbol) -> List[Mapping]:
        with self._client.safe_malloc(8) as p_result:
            p_result[0] = 0

            query = self._client.symbols.objc_getClass('NSMutableDictionary').objc_call('new')
            query.objc_call('setObject:forKey:', class_type[0],
                            self._client.symbols.kSecClass[0])
            query.objc_call('setObject:forKey:', self._client.symbols.kSecMatchLimitAll[0],
                            self._client.symbols.kSecMatchLimit[0])
            query.objc_call('setObject:forKey:', self._client.symbols.kCFBooleanTrue[0],
                            self._client.symbols.kSecReturnAttributes[0])
            query.objc_call('setObject:forKey:', self._client.symbols.kCFBooleanTrue[0],
                            self._client.symbols.kSecReturnRef[0])
            query.objc_call('setObject:forKey:', self._client.symbols.kCFBooleanTrue[0],
                            self._client.symbols.kSecReturnData[0])

            err = self._client.symbols.SecItemCopyMatching(query, p_result).c_int32
            if err != 0:
                raise BadReturnValueError(f'SecItemCopyMatching() returned: {err}')

            result = p_result[0]

            if result == 0:
                raise RpcPermissionError()

            # results contain a reference which isn't plist-serializable
            removal_key = self._client.cf('v_Ref')
            for i in range(result.objc_call('count')):
                result.objc_call('objectAtIndex:', i).objc_call('removeObjectForKey:', removal_key)
            return result.py()



import logging

from rpcclient.exceptions import RpcAppleScriptError

logger = logging.getLogger(__name__)


class AppleScript:
    def __init__(self, client):
        """
        :param client: rpcclient.macos.MacosClient
        """
        self._client = client

    def execute(self, script: str) -> None:
        with self._client.safe_malloc(8) as error:
            error[0] = 0
            apple_script = self._client.symbols.objc_getClass('NSAppleScript') \
                .objc_call('alloc').objc_call('initWithSource:', self._client.cf(script))
            apple_script.objc_call('executeAndReturnError:', error)
            if error[0]:
                raise RpcAppleScriptError(error[0].objc_call('description').py())
            apple_script.objc_call('release')

    def say(self, message: str, voice: str = None):
        script = f'say "{message}"'
        if voice is not None:
            script += f' using "{voice}"'
        self.execute(script)

    def beep(self, count: int) -> None:
        self.execute(f'beep {count}')



import plistlib
import struct
from typing import Mapping

import lief
from parameter_decorators import path_to_str

from rpcclient.darwin.consts import kSecCodeMagicEntitlement
from rpcclient.exceptions import NoEntitlementsError
from rpcclient.lief import Lief


class DarwinLief(Lief):
    @path_to_str('path')
    def get_entitlements(self, path: str) -> Mapping:
        with self._client.fs.open(path, 'r') as f:
            buf = f.read()
        parsed = lief.parse(buf)
        code_signature = buf[parsed.code_signature.data_offset:
                             parsed.code_signature.data_offset + parsed.code_signature.data_size]

        ent_magic = struct.pack('>I', kSecCodeMagicEntitlement)
        ent_magic_offset = code_signature.find(ent_magic)

        if ent_magic_offset == -1:
            raise NoEntitlementsError()

        ent_buf = code_signature[ent_magic_offset + len(ent_magic) + 4:]
        end_plist_magic = b'</plist>'
        ent_buf = ent_buf[:ent_buf.find(end_plist_magic) + len(end_plist_magic)]
        return plistlib.loads(ent_buf)



import logging

from rpcclient.structs.consts import RTLD_NOW

logger = logging.getLogger(__name__)


class Call:
    def __init__(self, client, controller, call):
        self._client = client
        self._controller = controller
        self._call = call

    def disconnect(self):
        """
        Disconnect the current call.
        """
        self._send_action('CXEndCallAction')

    def answer(self):
        """
        Answer the current call.
        """
        self._send_action('CXAnswerCallAction')

    def _send_action(self, action_name):
        action_class = self._client.symbols.objc_getClass(action_name)
        action = action_class.objc_call('alloc').objc_call('initWithCallUUID:', self._uuid)
        self._controller.objc_call('requestTransactionWithAction:completion:', action, self._client.get_dummy_block())

    @property
    def _uuid(self):
        return self._call.objc_call('UUID')


class Telephony:
    """
    Telephony utils.
    In order to access the real telephony, the API requires the entitlement "application-identifier" to be set to
    "com.apple.coretelephony".
    """

    def __init__(self, client):
        self._client = client
        self._load_callkit_library()
        self.cx_call_controller = self._client.symbols.objc_getClass('CXCallController').objc_call('new')
        self.cx_call_observer = self.cx_call_controller.objc_call('callObserver')
        self.ct_message_center = self._client.symbols.objc_getClass('CTMessageCenter').objc_call('sharedMessageCenter')

    def dial(self, number: str):
        """
        Start a call to a number. Use `current_call` to access the created call.
        :param number: Phone address to call to.
        """
        self._client.symbols.CTCallDial(self._client.cf(number))

    def send_sms(self, to_address: str, text: str, smsc: str = '1111'):
        """
        Send a SMS.
        :param to_address: Phone address to send to.
        :param text: Message text.
        :param smsc: Originator's short message service center address.
        """
        self.ct_message_center.objc_call(
            'sendSMSWithText:serviceCenter:toAddress:', self._client.cf(text), self._client.cf(smsc),
            self._client.cf(to_address)
        )

    @property
    def current_call(self) -> Call:
        """
        Return on object representing the current active call.
        """
        calls = self.cx_call_observer.objc_call('calls')
        call_count = calls.objc_call('count')

        call_list = []
        for i in range(call_count):
            call_list.append(calls.objc_call('objectAtIndex:', i))

        for call_id in range(call_count):
            call = call_list[call_id]
            if call.objc_call('hasEnded'):
                continue
            return Call(self._client, self.cx_call_controller, call)

    def _load_callkit_library(self):
        options = [
            '/System/Library/Frameworks/CallKit.framework/CallKit'
        ]
        for option in options:
            if self._client.dlopen(option, RTLD_NOW):
                return
        logger.warning('CallKit library isn\'t available')



from datetime import datetime

from rpcclient.darwin.structs import timeval
from rpcclient.exceptions import MissingLibraryError
from rpcclient.structs.consts import RTLD_NOW


class Time:
    def __init__(self, client):
        """
        :param rpcclient.darwin.client.DarwinClient client:
        """
        self._client = client
        self._load_core_time_framework()

    def _load_core_time_framework(self):
        if self._client.dlopen('/System/Library/PrivateFrameworks/CoreTime.framework/CoreTime', RTLD_NOW):
            return
        raise MissingLibraryError('failed to load CoreTime')

    def now(self) -> datetime:
        """ get current time """
        with self._client.safe_calloc(timeval.sizeof()) as current:
            self._client.symbols.gettimeofday(current, 0)
            time_of_day = timeval.parse(current.peek(timeval.sizeof()))
        return datetime.fromtimestamp(time_of_day.tv_sec + (time_of_day.tv_usec / (10 ** 6)))

    def set_current(self, new_time: datetime):
        """ set current time """
        self._client.symbols.TMSetAutomaticTimeZoneEnabled(0)
        with self._client.safe_calloc(timeval.sizeof()) as current:
            current.poke(timeval.build({'tv_sec': int(new_time.timestamp()), 'tv_usec': new_time.microsecond}))
            self._client.symbols.settimeofday(current, 0)

    def set_auto(self):
        """ opt-in automatic time settings """
        self._client.symbols.TMSetAutomaticTimeZoneEnabled(1)

    @property
    def is_set_automatically(self):
        """ tell is time settings are set to automatic """
        return bool(self._client.symbols.TMIsAutomaticTimeZoneEnabled())

    def boot_time(self) -> datetime:
        timestamp = timeval.parse(self._client.sysctl.get_by_name('kern.boottime')).tv_sec
        return datetime.fromtimestamp(timestamp)



import typing

from rpcclient.darwin.client import DarwinClient
from rpcclient.darwin.reports import Reports
from rpcclient.ios.accessibility import Accessibility
from rpcclient.ios.amfi import Amfi
from rpcclient.ios.backlight import Backlight
from rpcclient.ios.lockdown import Lockdown
from rpcclient.ios.mobile_gestalt import MobileGestalt
from rpcclient.ios.processes import IosProcesses
from rpcclient.ios.screen_capture import ScreenCapture
from rpcclient.ios.sprinboard import SpringBoard
from rpcclient.ios.telephony import Telephony
from rpcclient.ios.wifi import IosWifi

CRASH_REPORTS_DIR = 'Library/Logs/CrashReporter'


class IosClient(DarwinClient):
    def __init__(self, sock, sysname: str, arch, create_socket_cb: typing.Callable):
        super().__init__(sock, sysname, arch, create_socket_cb)
        self.backlight = Backlight(self)
        self.reports = Reports(self, CRASH_REPORTS_DIR)
        self.mobile_gestalt = MobileGestalt(self)
        self.processes = IosProcesses(self)
        self.lockdown = Lockdown(self)
        self.telephony = Telephony(self)
        self.screen_capture = ScreenCapture(self)
        self.accessibility = Accessibility(self)
        self.wifi = IosWifi(self)
        self.springboard = SpringBoard(self)
        self.amfi = Amfi(self)
        self._radio_preferences = self.symbols.objc_getClass('RadiosPreferences').objc_call('new')

    @property
    def roots(self) -> typing.List[str]:
        """ get a list of all accessible darwin roots when used for lookup of files/preferences/... """
        return super().roots + ['/var/mobile']

    @property
    def airplane_mode(self) -> bool:
        # use MobileGestalt for a more accurate result
        return self.mobile_gestalt.AirplaneMode

    @airplane_mode.setter
    def airplane_mode(self, value: bool):
        """ set whether the device should enter airplane mode (turns off baseband, bt, etc...) """
        self._radio_preferences.objc_call('setAirplaneMode:', value)
        self._radio_preferences.objc_call('synchronize')



import tempfile
from pathlib import Path
from stat import S_IMODE

import pytest

from rpcclient.darwin.consts import UF_IMMUTABLE
from rpcclient.exceptions import RpcFileNotFoundError, RpcPermissionError


def test_chown(client, tmp_path):
    file = (tmp_path / 'temp.txt')
    client.fs.touch(file)
    pre_chown_stat = client.fs.stat(file)
    client.fs.chown(file, 2, -1)
    post_chown_stat = client.fs.stat(file)
    assert post_chown_stat.st_uid == 2
    assert pre_chown_stat.st_gid == post_chown_stat.st_gid


def test_chmod(client, tmp_path):
    file = (tmp_path / 'temp.txt')
    client.fs.touch(file, mode=0o777)
    client.fs.chmod(file, 0o666)
    assert S_IMODE(client.fs.stat(file).st_mode) == 0o666


def test_remove(client, tmp_path):
    file = (tmp_path / 'temp.txt')
    client.fs.touch(file)
    client.fs.remove(file)
    assert not client.fs.accessible(file)


def test_mkdir(client, tmp_path):
    dir_ = (tmp_path / 'test_dir')
    client.fs.mkdir(dir_, 0o666)
    assert client.fs.accessible(dir_)


def test_chdir_pwd(client, tmp_path):
    client.fs.chdir(tmp_path)
    assert client.fs.pwd() == str(tmp_path)


def test_symlink(client, tmp_path):
    file = (tmp_path / 'temp.txt')
    symlink = (tmp_path / 'temp1.txt')
    client.fs.touch(file)
    client.fs.symlink(file, symlink)
    assert client.fs.accessible(symlink)
    assert client.fs.readlink(symlink) == str(file)


def test_realpath(client, tmp_path):
    assert client.fs.realpath(tmp_path / '././') == str(tmp_path)


def test_link(client, tmp_path):
    client.fs.write_file(tmp_path / 'temp.txt', b'hello')
    client.fs.link((tmp_path / 'temp.txt'), (tmp_path / 'temp1.txt'))
    assert client.fs.read_file(tmp_path / 'temp1.txt') == b'hello'


def test_listdir(client, tmp_path):
    assert not client.fs.listdir(tmp_path)
    client.fs.touch(tmp_path / 'temp.txt')
    assert client.fs.listdir(tmp_path) == ['temp.txt']


def test_listdir_non_exists(client):
    with pytest.raises(RpcFileNotFoundError):
        client.fs.listdir('/non_exists_path')


def test_push_expand_path(client, tmp_path):
    with tempfile.TemporaryDirectory() as temp_dir:
        local_file = Path(temp_dir) / 'local'
        local_file.touch()
        remote_file = tmp_path / 'temp.txt'
        client.fs.push(local_file, remote_file)
        assert remote_file.exists()


def test_pull_expand_path(client, tmp_path):
    with tempfile.TemporaryDirectory() as temp_dir:
        remote_file = tmp_path / 'temp.txt'
        remote_file.touch()
        local_file = Path(temp_dir) / 'local'
        client.fs.pull(remote_file, local_file)
        assert local_file.exists()


@pytest.mark.parametrize('file_size', [1024, 10240, 102400])
def test_push_pull_with_different_sizes(client, tmp_path, file_size):
    assert not client.fs.listdir(tmp_path)
    local = Path('/tmp/temp.bin')
    local_pull = Path('/tmp/temp2.bin')
    remote = tmp_path / 'temp.bin'

    with open(local, 'wb') as f:
        f.write(b'\0' * file_size)

    client.fs.push(local, remote)

    assert client.fs.lstat(remote).st_size == file_size
    client.fs.pull(remote, local_pull)
    assert local_pull.stat().st_size == file_size
    local.unlink(missing_ok=True)
    local_pull.unlink(missing_ok=True)


def test_pull(client, tmp_path):
    client.fs.touch(tmp_path / 'a')
    with tempfile.TemporaryDirectory() as local_dir:
        local_dir = Path(local_dir)
        client.fs.pull(tmp_path / 'a', local_dir)
        assert (local_dir / 'a').exists()
    with tempfile.TemporaryDirectory() as local_dir:
        local_dir = Path(local_dir)
        client.fs.pull(tmp_path / 'a', local_dir / 'a')
        assert (local_dir / 'a').exists()

    client.fs.mkdir(tmp_path / 'b')
    with tempfile.TemporaryDirectory() as local_dir:
        local_dir = Path(local_dir)
        client.fs.pull(tmp_path / 'b', local_dir, recursive=True)
        assert (local_dir / 'b').exists()
        client.fs.pull(tmp_path / 'b', local_dir / 'b', recursive=True)
        assert (local_dir / 'b' / 'b').exists()


def test_push_pull_dir(client, tmp_path: Path):
    (tmp_path / 'a').touch()
    with tempfile.TemporaryDirectory() as local_dir:
        local_dir = Path(local_dir)
        client.fs.pull(tmp_path, local_dir, recursive=True)
        assert (local_dir / tmp_path.name / 'a').exists()
    with tempfile.TemporaryDirectory() as local_dir:
        local_dir = Path(local_dir)
        (local_dir / 'b').touch()
        client.fs.push(local_dir, tmp_path, recursive=True)
        a = tmp_path / local_dir.name / 'b'
        assert a.exists()


def test_scandir_sanity(client, tmp_path):
    entries = [e for e in client.fs.scandir(tmp_path)]
    assert not entries
    client.fs.write_file(tmp_path / 'temp.txt', b'hello')
    entries = [e for e in client.fs.scandir(tmp_path)]
    assert len(entries) == 1
    assert entries[0].name == 'temp.txt'
    assert entries[0].path == str(tmp_path / 'temp.txt')
    assert entries[0].is_file()
    assert not entries[0].is_dir()
    assert not entries[0].is_symlink()


def test_stat_sanity(client, tmp_path):
    file = tmp_path / 'temp.txt'
    client.fs.write_file(file, 'h' * 0x10000)
    client_stat = client.fs.stat(file)
    path_stat = client.fs.stat(file)
    assert path_stat.st_dev == client_stat.st_dev
    assert path_stat.st_ino == client_stat.st_ino
    assert path_stat.st_mode == client_stat.st_mode
    assert path_stat.st_nlink == client_stat.st_nlink
    assert path_stat.st_uid == client_stat.st_uid
    assert path_stat.st_gid == client_stat.st_gid
    assert path_stat.st_rdev == client_stat.st_rdev
    assert path_stat.st_atime == client_stat.st_atime
    assert path_stat.st_mtime == client_stat.st_mtime
    assert path_stat.st_ctime == client_stat.st_ctime
    assert path_stat.st_size == client_stat.st_size
    assert path_stat.st_blocks == client_stat.st_blocks
    assert path_stat.st_blksize == client_stat.st_blksize
    assert path_stat.st_flags == client_stat.st_flags
    assert path_stat.st_gen == client_stat.st_gen


def test_walk(client, tmp_path):
    client.fs.mkdir(tmp_path / 'dir_a')
    client.fs.touch(tmp_path / 'dir_a' / 'a1.txt')
    client.fs.touch(tmp_path / 'dir_a' / 'a2.txt')
    client.fs.mkdir(tmp_path / 'dir_b')
    client.fs.touch(tmp_path / 'dir_b' / 'b1.txt')
    client.fs.touch(tmp_path / 'dir_b' / 'b2.txt')

    assert list(client.fs.walk(tmp_path)) == [
        (f'{tmp_path}', ['dir_b', 'dir_a'], []),
        (f'{tmp_path}/dir_b', [], ['b2.txt', 'b1.txt']),
        (f'{tmp_path}/dir_a', [], ['a1.txt', 'a2.txt'])
    ]


@pytest.mark.darwin
def test_xattr(client, tmp_path):
    client.fs.setxattr(tmp_path, 'KEY', b'VALUE')
    assert client.fs.getxattr(tmp_path, 'KEY') == b'VALUE'
    assert client.fs.listxattr(tmp_path) == ['KEY']
    assert client.fs.dictxattr(tmp_path) == {'KEY': b'VALUE'}
    client.fs.removexattr(tmp_path, 'KEY')
    assert client.fs.listxattr(tmp_path) == []


@pytest.mark.darwin
def test_chflags(client, tmp_path):
    # create temporary file
    file = tmp_path / 'file'
    with client.fs.open(file, 'w'):
        pass
    # make it immutable
    client.fs.chflags(file, UF_IMMUTABLE)
    # verify we cannot remove it
    with pytest.raises(RpcPermissionError):
        client.fs.remove(file)
    # restore its permissions
    client.fs.chflags(file, 0)
    # verify removal succeeds
    client.fs.remove(file)



from rpcclient.exceptions import BadReturnValueError
from rpcclient.structs.consts import SIGTERM


class Processes:
    def __init__(self, client):
        """
        process manager
        :param rpcclient.darwin_client.Client client: client
        """
        self._client = client

    def kill(self, pid: int, sig: int = SIGTERM):
        """ kill(pid, sig) at remote. read man for more details. """
        if 0 != self._client.symbols.kill(pid, sig):
            raise BadReturnValueError(f'kill({pid}, {sig}) failed ({self._client.last_error})')

    def waitpid(self, pid: int, flags: int = 0):
        """ waitpid(pid, sig) at remote. read man for more details. """
        with self._client.safe_malloc(8) as stat_loc:
            err = self._client.symbols.waitpid(pid, stat_loc, flags).c_int64
            if err == -1:
                raise BadReturnValueError(f'waitpid(): returned {err} ({self._client.last_error})')
            return stat_loc[0]



from typing import Mapping

from rpcclient.allocated import Allocated
from rpcclient.darwin.consts import MACH_PORT_NULL, kCFAllocatorDefault, kIOServicePlane
from rpcclient.darwin.structs import io_name_t, io_object_t, mach_port_t
from rpcclient.exceptions import BadReturnValueError, RpcClientException


class IOService(Allocated):
    """ representation of a remote IOService """

    def __init__(self, client, service):
        super().__init__()
        self._client = client
        self._service = service

    @property
    def name(self) -> str:
        with self._client.safe_malloc(io_name_t.sizeof()) as name:
            if self._client.symbols.IORegistryEntryGetName(self._service, name):
                raise BadReturnValueError('IORegistryEntryGetName failed')
            return name.peek_str()

    @property
    def properties(self) -> Mapping:
        with self._client.safe_malloc(8) as p_properties:
            if self._client.symbols.IORegistryEntryCreateCFProperties(self._service, p_properties,
                                                                      kCFAllocatorDefault, 0):
                raise BadReturnValueError('IORegistryEntryCreateCFProperties failed')
            return p_properties[0].py()

    def __iter__(self):
        with self._client.safe_malloc(io_object_t.sizeof()) as p_child_iter:
            if self._client.symbols.IORegistryEntryGetChildIterator(self._service, kIOServicePlane, p_child_iter):
                raise BadReturnValueError('IORegistryEntryGetChildIterator failed')
            child_iter = p_child_iter[0]

        while True:
            child = self._client.symbols.IOIteratorNext(child_iter)
            if not child:
                break
            s = IOService(self._client, child)
            yield s

    def set(self, properties: Mapping):
        self._client.symbols.IORegistryEntrySetCFProperties(self._service, self._client.cf(properties))

    def get(self, key: str):
        return self._client.symbols.IORegistryEntryCreateCFProperty(self._service, self._client.cf(key),
                                                                    kCFAllocatorDefault, 0).py()

    def _deallocate(self):
        self._client.symbols.IOObjectRelease(self._service)

    def __repr__(self):
        return f'<{self.__class__.__name__} NAME:{self.name}>'


class BacklightControlService(IOService):
    @property
    def display_parameters(self) -> Mapping:
        return self.get('IODisplayParameters')

    @property
    def brightness(self) -> int:
        return self.display_parameters['brightness']['value']

    @brightness.setter
    def brightness(self, value: int):
        self.set({'EnableBacklight': bool(value)})
        self.set({'brightness': value})


class PowerSourceService(IOService):
    @property
    def battery_voltage(self) -> int:
        return self.get('AppleRawBatteryVoltage')

    @property
    def charging(self) -> bool:
        return self.get('IsCharging')

    @charging.setter
    def charging(self, value: bool):
        self.set({'IsCharging': value, 'ExternalConnected': value})

    @property
    def external_connected(self) -> bool:
        return self.get('ExternalConnected')

    @external_connected.setter
    def external_connected(self, value: bool):
        self.set({'ExternalConnected': value})

    @property
    def current_capacity(self) -> int:
        return self.get('CurrentCapacity')

    @current_capacity.setter
    def current_capacity(self, value: int):
        self.set({'CurrentCapacity': value})

    @property
    def at_warn_level(self) -> bool:
        return self.get('AtWarnLevel')

    @at_warn_level.setter
    def at_warn_level(self, value: bool):
        self.set({'AtWarnLevel': value})

    @property
    def time_remaining(self) -> int:
        return self.get('TimeRemaining')

    @property
    def temperature(self) -> int:
        return self.get('Temperature')

    @temperature.setter
    def temperature(self, value: int):
        self.set({'Temperature': value})


class IORegistry:
    """
    IORegistry utils
    https://developer.apple.com/library/archive/documentation/DeviceDrivers/Conceptual/IOKitFundamentals/TheRegistry/TheRegistry.html
    """

    def __init__(self, client):
        """
        :param rpcclient.darwin.client.DarwinClient client:
        """
        self._client = client

    @property
    def backlight_control(self) -> BacklightControlService:
        service = self._client.symbols.IOServiceGetMatchingService(
            0, self._client.cf({'IOPropertyMatch': {'backlight-control': True}}))
        if not service:
            raise RpcClientException('IOServiceGetMatchingService failed')
        return BacklightControlService(self._client, service)

    @property
    def power_source(self) -> PowerSourceService:
        service = self._client.symbols.IOServiceGetMatchingService(
            0, self._client.symbols.IOServiceMatching('IOPMPowerSource'))
        if not service:
            raise RpcClientException('IOServiceGetMatchingService failed')
        return PowerSourceService(self._client, service)

    @property
    def root(self) -> IOService:
        with self._client.safe_malloc(mach_port_t.sizeof()) as p_master_port:
            self._client.symbols.IOMasterPort(MACH_PORT_NULL, p_master_port)
            return IOService(self._client, self._client.symbols.IORegistryGetRootEntry(p_master_port[0]))



import pytest

from rpcclient.exceptions import RpcPermissionError

pytestmark = pytest.mark.darwin


def test_query_certificates(client):
    try:
        assert len(client.keychain.query_certificates()) > 1
    except RpcPermissionError:
        pass


def test_query_keys(client):
    try:
        assert len(client.keychain.query_keys()) > 1
    except RpcPermissionError:
        pass



import pytest

from rpcclient.exceptions import ArgumentError


def test_peek(client):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_malloc(0x100) as peekable:
        client.peek(peekable, 0x100)


def test_poke(client):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_malloc(0x100) as peekable:
        client.poke(peekable, b'a' * 0x100)


@pytest.mark.parametrize('params', [
    ([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
    ([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15])
])
def test_16args(client, params):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_malloc(0x100) as peekable:
        client.symbols.test_16args(peekable, *params)
        for i in range(len(params)):
            assert peekable[i] == params[i]


@pytest.mark.parametrize('va_list_index,params,expected', [
    (2, ['%f', 15.5], '15.500000'),
    (2, ['%f %f', 15.5, 7.3], '15.500000 7.300000'),
    (2, ['%f %f %d', 15.5, 7.3, 42], '15.500000 7.300000 42'),
    (2, ['%f %f %d %d', 15.5, 7.3, 1, 2], '15.500000 7.300000 1 2'),
    (2, ['%f %f %d %d %s', 15.5, 7.3, 1, 2, 'test'], '15.500000 7.300000 1 2 test'),
    (2, ['%f %f %d %d %s %s', 15.5, 7.3, 1, 2, 'test', 'test2'], '15.500000 7.300000 1 2 test test2'),
])
@pytest.mark.arm
def test_va_list_call(client, va_list_index, params, expected):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_malloc(0x100) as peekable:
        client.symbols.sprintf(peekable, *params, va_list_index=va_list_index)
        assert expected == peekable.peek_str()


@pytest.mark.arm
def test_floating_point_call(client):
    """
    :param rpcclient.client.Client client:
    """
    assert client.symbols.sqrt(16.0, return_float64=True) == 4.0


def test_peek_invalid_address(client):
    """
    :param rpcclient.client.Client client:
    """
    # Server config: -DSAFE_WRITE raises ArgumentError for invalid address; else, ConnectionError.
    with pytest.raises((ArgumentError, ConnectionError)):
        client.peek(0, 0x10)


def test_poke_invalid_address(client):
    """
    :param rpcclient.client.Client client:
    """
    # Server config: -DSAFE_WRITE raises ArgumentError for invalid address; else, ConnectionError.
    with pytest.raises((ArgumentError, ConnectionError)):
        client.poke(0, b'a')


def test_get_dummy_block(client):
    """
    :param rpcclient.client.Client client:
    """
    assert client.get_dummy_block()[0] == client.symbols._NSConcreteGlobalBlock


def test_listdir(client):
    """
    :param rpcclient.client.Client client:
    """
    entries = client.listdir('/')
    assert entries[0].d_name == '.'
    assert entries[1].d_name == '..'


def test_calloc(client):
    """
    :param rpcclient.client.Client client:
    """
    with client.safe_calloc(0x100) as zeros:
        assert client.peek(zeros, 0x100) == b'\x00' * 0x100


def test_env_get_set(client):
    """
    :param rpcclient.client.Client client:
    """
    client.setenv('TEST', 'test')
    assert client.getenv('TEST') == 'test'


def test_environ(client):
    """
    :param rpcclient.client.Client client:
    """
    assert len(client.environ) > 0
    for e in client.environ:
        assert '=' in e


def test_reconnect(client):
    """
    :param rpcclient.client.Client client:
    """
    test_listdir(client)
    client.reconnect()
    test_listdir(client)



import time
from threading import Thread


def test_same_socket_different_threads(client):
    # get an expected result once on main thread
    expected_result = client.fs.listdir('/')

    # global to tell threads when to exit
    should_exit = False

    def listdir_thread(client):
        while not should_exit:
            assert client.fs.listdir('/') == expected_result
        return 0

    # launch the two threads
    t1 = Thread(target=listdir_thread, args=(client,))
    t2 = Thread(target=listdir_thread, args=(client,))

    t1.start()
    t2.start()

    # wait 10 seconds
    time.sleep(10)

    # tell threads they
    should_exit = True

    t1.join()
    t2.join()
